import logging
import time
import typing as t

from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.bioshake_pb2 import Command, Config

from .driver import BioshakeDriver
import argparse 

class BioShakeServer(ToolServer):
    toolType = "bioshake"

    def __init__(self) -> None:
        super().__init__()
        self.driver: t.Optional[BioshakeDriver] = None

    def _configure(self, config: Config) -> None:
        self.config = config
        if self.driver:
            logging.info("Bioshake driver already exists, reconfiguring...")
            self.driver.port = self.config.com_port
        else:
            self.driver = BioshakeDriver(port=self.config.com_port)
        self.driver.connect()

    def Grip(self, params: Command.Grip) -> None:
        if not self.driver:
            raise Exception("Bioshake driver not connected")
        self.driver.set_elm_lock_pos()

    def Ungrip(self, params: Command.Ungrip) -> None:
        if not self.driver:
            raise Exception("Bioshake driver not connected")
        self.driver.set_elm_unlock_pos()

    def Home(self, params: Command.Home) -> None:
        if not self.driver:
            raise Exception("Bioshake driver not connected")
        self.driver.shake_go_home()

    def StartShake(self, params: Command.StartShake) -> None:
        if not self.driver:
            raise Exception("Bioshake driver not connected")
        self.driver.set_shake_target_speed(rpm=params.speed)
        if params.duration > 0:
            # Note, this is not blocking meaning that command will return
            # And next command will be executed immediately
            # This enables shaking to happen asynchronously which enables
            # stop commands to be issued, but also means that workcell can
            # do other things while shaking is happening
            # However, this also means that when we want to wait for
            # shaking to finish, we need to use WaitForShakeToFinish
            self.driver.shake_on_with_runtime(seconds=params.duration)
        else:
            self.driver.shake_on()

    def StopShake(self, params: Command.StopShake) -> None:
        if not self.driver:
            raise Exception("Bioshake driver not connected")
        self.driver.shake_off()
        time.sleep(4)

    def WaitForShakeToFinish(self, params: Command.WaitForShakeToFinish) -> None:
        """Note, this is a blocking command that waits for shake to finish
        for a maximum of params.timeout seconds. If shake is not finished
        within that time, it will return None. If it is desired for shake to
        end, then a stop shake command will need to be sent. Alternatively,
        the shake will stop once the time originally given to it has elapsed.
        """
        if not self.driver:
            raise Exception("Bioshake driver not connected")
        start_time = time.time()
        while time.time() - start_time < params.timeout:
            remaining_time = self.driver.get_shake_remaining_time()
            shake_state = self.driver.get_shake_state_as_string()
            logging.info(
                f"Current status: {shake_state}... Shake remaining time: {remaining_time}"
            )
            if remaining_time == 0 or shake_state in ["STOP", "ESTOP"]:
                logging.info("Shake finished")
                return None
            else:
                time.sleep(5)
        logging.warning("Timeout reached shake not finished")
        return None

    def Reset(self, params: Command.Reset) -> None:
        if not self.driver:
            raise Exception("Bioshake driver not connected")
        self.driver.reset_device()

    def EstimateGrip(self, params: Command.Grip) -> int:
        return 1

    def EstimateUngrip(self, params: Command.Ungrip) -> int:
        return 1

    def EstimateHome(self, params: Command.Home) -> int:
        return 1

    def EstimateStartShake(self, params: Command.StartShake) -> int:
        return 1

    def EstimateStopShake(self, params: Command.StopShake) -> int:
        return 1

    def EstimateReset(self, params: Command.Reset) -> int:
        return 1

    def EstimateWaitForShakeToFinish(self, params: Command.WaitForShakeToFinish) -> int:
        return 1


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
        raise RuntimeWarning("Port must be provided...")
    serve(BioShakeServer(), str(args.port))



import logging
from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.toolbox_pb2 import Command, Config
from tools.app_config import Config as GlobalConfig
from .data import Data 
from tools.grpc_interfaces.tool_base_pb2 import ExecuteCommandReply
import argparse 
from tools.toolbox.slack import Slack 
from google.protobuf.struct_pb2 import Struct
from tools.grpc_interfaces.tool_base_pb2 import  SUCCESS, ERROR_FROM_TOOL
from tools.grpc_interfaces import tool_base_pb2
from tools.toolbox.python_subprocess import run_python_script

class ToolBoxServer(ToolServer):
     toolType = "toolbox"

     def __init__(self) -> None:
          super().__init__()
          self.app_config = GlobalConfig()
          self.app_config.load_workcell_config()
          self.slack = Slack(self.app_config)
          self.setStatus(tool_base_pb2.READY)

     def _configure(self, request:Config) -> None:
          return
     
     # def SlackMessage(self, params:Command.SlackMessage) -> None:
     #      if self.app_config.app_config.slack_workcell_channel:
     #           self.slack.slack_message(message=params.message, recipient=self.app_config.app_config.slack_workcell_channel)

     def LogMediaExchange(self, params:Command.LogMediaExchange) -> None: 
         #Data.log_media_exchange(params.source_barcode, params.destination_name, params.destination_barcode,params.source_wells, params.percent_exchange,params.new_tips)
          return None 
     
     def GetWorkcells(self, params:Command.GetWorkcells) -> ExecuteCommandReply:
          s  = Struct()
          response = ExecuteCommandReply()
          response.return_reply = True
          response.response = SUCCESS
          try:
               workcells = Data.get_workcells()
               if workcells:
                    s.update({'workcells':workcells})
               else:
                   s.update({'workcells':[]})
               response.meta_data.CopyFrom(s)
          except Exception as exc:
               logging.exception(exc)
               response.response = ERROR_FROM_TOOL

          return response
     
     def GetLiconicSensorData(self, params:Command.GetLiconicSensorData) -> None:
          return None 
          # s  = Struct()
          # response = ExecuteCommandReply()
          # response.return_reply = True
          # response.response = SUCCESS
          # try:
          #      data = Data.get_liconic_sensor_data(params.tool_id, params.date)
          #      if data:
          #           s.update(data)
          #      else:
          #           s.update({'times':[],'co2_values':[]})
          #      response.meta_data.CopyFrom(s)
          # except Exception as exc:
          #      logging.exception(exc)
          #      response.response = ERROR_FROM_TOOL

          # return response

     
     def GetOT2ImagesByDate(self, params:Command.GetOT2ImagesByDate) -> None:
          return None 
          #s  = Struct()
          # response = ExecuteCommandReply()
          # response.return_reply = True
          # response.response = SUCCESS
          # try:
          #      data = Data.get_ot2_images_by_date(params.date)
          #      if data:
          #           logging.info("Result should be "+str(data))
          #           s.update({'images':data})
          #      else:
          #           s.update({'images':[]})
          #      response.meta_data.CopyFrom(s)
          # except Exception as exc:
          #      logging.exception(exc)
          #      response.response = ERROR_FROM_TOOL
          # return response
     
     def GetOT2ImageBytes(self, params:Command.GetOT2ImageBytes) -> None:
          return None 
          # s  = Struct()
          # response = ExecuteCommandReply()
          # response.return_reply = True
          # response.response = SUCCESS
          # try:
          #      data = Data.get_ot2_image_bites_by_file_name(params.date, params.image_file)
          #      if data:
          #           s.update(data)
          #      else:
          #          s.update({'created_on':'','img_bytes':''})
          #      response.meta_data.CopyFrom(s)
          # except Exception as exc:
          #      logging.exception(exc)
          #      response.response = ERROR_FROM_TOOL
          # return response
     
     def RunScript(self, params:Command.RunScript) -> ExecuteCommandReply:
          s  = Struct()
          response = ExecuteCommandReply()
          response.return_reply = True
          response.response = SUCCESS
          try:
               result = run_python_script(params.script_content,blocking=True)
               logging.info(f"Script result is {result}")
               if response:
                    s.update({'response':result})
               else:
                    s.update({'response':''})
               response.meta_data.CopyFrom(s)
          except Exception as exc:
               logging.exception(exc)
               response.response = ERROR_FROM_TOOL
               response.error_message = str(exc)
          return response
     
     # def SendSlackAlert(self, params:Command.SendSlackAlert) -> None:
     #      if self.app_config.app_config.slack_error_channel:
     #           self.slack.send_alert_slack(params.workcell, params.tool, params.protocol, params.error_message, self.app_config.app_config.slack_error_channel)
     
     def ClearLastSlackAlert(self, params:Command.ClearLastSlackAlert) -> None:
          # if self.app_config.app_config.slack_error_channel:
          #      self.slack.clear_last_error(self.app_config.app_config.slack_error_channel)
          return None 
     
     def GetLogMediaExchangeByDate(self, params:Command.GetLogMediaExchangeByDate) -> None:
          return None 
          # s  = Struct()
          # response = ExecuteCommandReply()
          # response.return_reply = True
          # response.response = SUCCESS
          # try:
          #      data = Data.get_media_exchange_logs_by_date(params.date)
          #      if data:
          #           logging.info(f"Data is {data}")
          #           s.update({'data':data})
          #      else:
          #          s.update({})
          #      response.meta_data.CopyFrom(s)
          # except Exception as exc:
          #      logging.exception(exc)
          #      response.response = ERROR_FROM_TOOL
          # return response
     
     def ValidateFolder(self, params:Command.ValidateFolder) -> ExecuteCommandReply:
          logging.info("Running validate folder in the server")
          s  = Struct()
          response = ExecuteCommandReply()
          response.return_reply = True
          response.response = SUCCESS
          try:
               data = Data.validate_folder(params.folder_path)
               if data:
                    s.update({'result':True})
               else:
                   s.update({'result':False})
               response.meta_data.CopyFrom(s)
          except Exception as exc:
               logging.exception(exc)
               response.response = ERROR_FROM_TOOL
          return response
             
if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
         raise RuntimeWarning("Port must be provided...")
    serve(ToolBoxServer(), str(args.port))



import logging

from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.alps3000_pb2 import Command, Config
from .driver import ALPS3000Driver
import argparse
import time 

class ALPS3000Server(ToolServer):
    toolType = "alps3000"

    def __init__(self) -> None:
        super().__init__()
        self.driver: ALPS3000Driver

    def _configure(self, request: Config) -> None:
        logging.info("Initializing ALPS 3000...")
        self.config = request
        self.driver = ALPS3000Driver(profile=request.profile, port=request.com_port)
        return self.driver.initialize()

    def GetInstrumentStatus(self, params: Command.GetInstrumentStatus) -> None:
        return self.driver.get_status()

    def SealPlate(self, params: Command.SealPlate) -> None:
        self.driver.seal_plate()
        time.sleep(12)

    def GetError(self, params: Command.GetError) -> None:
        self.driver.get_error()

    def SetTemperature(self, params: Command.SetTemperature) -> None:
        self.driver.set_sealing_temperature(params.temperature)

    def SetSealTime(self, params: Command.SetSealTime) -> None:
        self.driver.set_sealing_time(params.seal_time)

    def GetTemperatureSetpoint(self, params: Command.GetTemperatureSetpoint) -> None:
        self.driver.get_sealing_temperature_setpoint()

    def GetSealingTime(self, params: Command.GetSealingTime) -> None:
        self.driver.get_sealing_time()

    def GetTemperatureActual(self, params: Command.GetTemperatureActual) -> None:
        self.driver.get_sealing_temperature_actual()

    def EstimateSealPlate(self, params: Command.SealPlate) -> int:
        return 1
    
    def EstimateGetError(self, params: Command.GetError) -> int:
        return 1
    
    def EstimateSetTemperature(self, params: Command.SetTemperature) -> int:
        return 1
    


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
        raise RuntimeWarning("Port must be provided...")
    serve(ALPS3000Server(), str(args.port))



import logging
import os

from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.multidrop_pb2 import  Config

class MultidropServer(ToolServer):
    toolType = "multidrop"
    #driver: HiGCentrifugeDriver
    config: Config

    def __init__(self) -> None:
        super().__init__()
    
    def _configure(self, request:Config) -> None:
        logging.info("Initializing combidrop")
        self.config = request


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    serve(MultidropServer(), os.environ.get("PORT", "4600"))



import logging

from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.dataman70_pb2 import Command, Config

from .driver import Dataman70Driver
import argparse

class Dataman70Server(ToolServer):
  toolType = "dataman70"
  driver: Dataman70Driver
  config: Config

  def __init__(self) -> None:
    super().__init__()
  
  def _configure(self, config: Config) -> None:
    self.config = config
    if self.driver:
      self.driver.close()
    self.driver = Dataman70Driver(self.config.com_port)
    self.driver.barcode_listeners.append(self._on_barcode)
    self.driver.power_on()
  
  def _on_barcode(self, barcode: str) -> None:
    logging.debug(f"Barcode detected! {barcode}")
    self.driver.live = False
  
  def Reset(self, params: Command.Reset) -> None:
    self.driver.power_off()
    self.driver.power_on()

  def AssertBarcode(self, params: Command.AssertBarcode) -> None:
    self.driver.live = True
    barcode = self.driver.read_barcode()
    # Dataman70Driver adds a leading 0 to the barcode because it outputs EAN-13 barcodes
    # barcodes are UPC-A, which are 12 digits long
    # We add a leading 0 to the UPC-A barcode to correctly match the barcode scanned
    logging.debug(f"Barcode scanned: {barcode}")
    if barcode != '0'+params.barcode:
      raise Exception(f"Expected barcode {params.barcode}, got {barcode}")
    return None
  
  def EstimateReset(self, params: Command.Reset) -> int:
    return 5
  
  def EstimateAssertBarcode(self, params: Command.AssertBarcode) -> int:
    return 5

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
        raise RuntimeWarning("Port must be provided...")
    serve(Dataman70Server(), str(args.port))



from slack_sdk import WebClient
from dotenv import load_dotenv
import os 
import requests
from tools.base_server import ABCToolDriver

load_dotenv(
    dotenv_path=os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "..",
        ".env.local",
    )
)

class Messages(ABCToolDriver):
    def __init__(self) -> None:
        self.slack_channel: str = os.getenv("ACTIVE_CULTURE_CHANNEL") # type: ignore
        self.slack_token: str = os.getenv("WORKCELL_BOT_TOKEN") # type: ignore


    def slack_message(self, message: str, channel:str) -> None:
        try:
            client = WebClient(token=self.slack_token)
            if not self.slack_channel:
                # throw error
                raise Exception("No slack channel specified")
            client.chat_postMessage(channel=self.slack_channel, text=message)

        except requests.exceptions.RequestException as e:
            raise Exception(
                f"Unable to post slack message to {self.slack_channel}"
            ) from e
    
    def send_email(self, message:str) -> None:
        return None
    
    


import serial
import time
import logging
from tools.base_server import ABCToolDriver

def try_ascii_decode(data: bytes) -> str:
    try:
        return data.decode("ascii")
    except Exception as e:
        raise serial.SerialException(f"ASCII decode error: {data!r} - {e}")

def serial_read(serial_port: serial.Serial) -> str:
    reply = serial_port.read_until(expected=b"\r\n")
    reply_string = try_ascii_decode(reply)
    if reply_string == "":
        logging.warning("XPeel returned empty response")
        return ""
    return reply_string.strip()

class XPeelDriver(ABCToolDriver):
    def __init__(self,com_port: str) -> None:
        self.serial_port = serial.Serial(
            com_port, 9600, timeout=1, parity=serial.PARITY_NONE,
            stopbits=serial.STOPBITS_ONE, bytesize=serial.EIGHTBITS
        )
        self.write("*stat")
        self.wait_for_ready()

    def write(self, command: str) -> None:
        self.serial_port.write((command + "\r\n").encode("ascii"))

    def read(self, strict: bool = True, max_attempts: int = 10) -> str:
        attempt = 0
        while attempt < max_attempts:
            reply_string = serial_read(self.serial_port)
            if reply_string == "":
                attempt += 1
                logging.info(f"Retrying...Attempt {attempt}")
                continue
            return reply_string
        if strict:
            raise serial.SerialException("No valid data received, may have disconnected.")
        return ''

    def wait_for_ready(self) -> None:
        """ Wait until the device sends a ready message indicating it can accept new commands. """
        while True:
            response = self.read(strict=False)
            logging.info(f"Received response from xpeel: {response}")
            if response.startswith("*ready"):
                return
            time.sleep(0.1)  # short delay to prevent flooding the serial port

    def remove_seal(self) -> None:
        """ Perform a deseal operation with specific parameters. """
        self.write("*xpeel:41")
        self.wait_for_ready()

    def check_status(self) -> str:
        """ Request current status and error codes. """
        self.write("*stat")
        return self.read()

    def reset(self) -> None:
        """ Reset the device to a known good state. """
        self.write("*reset")
        self.wait_for_ready()

    def restart(self) -> None:
        """ Restart the device. """
        self.write("*restart")
        self.wait_for_ready()

    def check_tape_remaining(self) -> str:
        """ Check how much tape is left on the spools. """
        self.write("*tapeleft")
        return self.read()

    def move_conveyor_in(self) -> None:
        """ Move the conveyor in to the default begin peel position. """
        self.write("*movein")
        self.wait_for_ready()

    def move_conveyor_out(self) -> None:
        """ Move the conveyor out. """
        self.write("*moveout")
        self.wait_for_ready()

    def move_elevator_up(self) -> None:
        """ Move the elevator up. """
        self.write("*moveup")
        self.wait_for_ready()

    def move_elevator_down(self) -> None:
        """ Move the elevator down. """
        self.write("*movedown")
        self.wait_for_ready()

    def move_spool(self) -> None:
        """ Advance the spool by 10mm of tape. """
        self.write("*movespool")
        self.wait_for_ready()

    def setup_plate_check(self, enable: bool) -> None:
        """ Enable or disable plate check feature. """
        on_off = 'y' if enable else 'n'
        self.write(f"*platecheck:{on_off}")
        self.wait_for_ready()

    def close(self) -> None:
        self.serial_port.close()



from slack_sdk import WebClient
import json
import os 
import logging
import requests
from typing import Any
from tools.app_config import Config
from typing import Optional 
import time 

slack_active = True

class Slack():
    def __init__(self, config:Config) -> None:
        self.client = WebClient(token="")

    def slack_message(self, message:Optional[str], recipient:str) -> None:
        try:
            self.client.chat_postMessage(
                channel=recipient,
                text=message
            )
        except requests.exceptions.RequestException as e:
            raise Exception(
                f"Unable to post slack message to {recipient}"
            ) from e

    def slack_block_message(self, message:Any,attachments:Any, recipient:str) -> None:
        try:
            self.client.chat_postMessage(
                channel=recipient,
                text=message,
                attachments=attachments
            )
        except requests.exceptions.RequestException as e:
            raise Exception(
                f"Unable to post slack message to {recipient}"
            ) from e
        
    def create_alert_message(self,title:str, workcell:str, tool:str, protocol:str, error:str) -> Any:
        template_path = os.path.join(os.path.dirname(__file__),"slack_templates","error_template.json")
        with open(template_path) as f:
            template_json = json.load(f)
        
        template_string = json.dumps(template_json)
        template_string = template_string.replace("<title_holder>", title)
        template_string = template_string.replace("<workcell_holder>", workcell)

        timestamp = str(time.strftime("%Y-%m-%d %H:%M:%S"))
        template_string = template_string.replace("<time_holder>", timestamp)
        template_string = template_string.replace('"<time_holder_int>"', str(time.time()))
        template_string = template_string.replace("<tool_holder>", tool)
        template_string = template_string.replace("<protocol_holder>", protocol)
        template_string = template_string.replace("<error_placeholder>", error)
        template_string = template_string.replace("<color_placeholder>", "#FF0000")

        return json.loads(template_string)
    
    def update_alert(self, workcell:str, tool:str, protocol:str, error:str, channel_id:str,status_update:str, message_id:str) -> None:
        msg =self.create_alert_message(title=status_update, workcell=workcell,tool=tool,protocol=protocol,error=error)
        self.client.chat_update(channel=channel_id,ts=message_id,text="",attachments=msg)
    
    def clear_last_error(self, channel_id:str) -> None:
        last_messages = self.get_last_messages(channel_id=channel_id)
        for msg in last_messages:
            #print(f"Message is {msg}")
            if msg:
                if 'attachments' in msg:
                    msg_resp = msg['attachments'][0]
                    if 'title' in msg_resp:
                        print(f"Title is {msg_resp['title']}")
                        if msg_resp['title'] == 'Workcell has errored!':
                            msg_resp['title'] = 'Error is resolved!'
                            msg_resp['color'] = '#36a64f'
                        #If the last message is not in error state, exit.
                        else:
                            continue
                    if 'fields' in msg_resp:
                        new_fields = msg_resp['fields']
                        for i in range(len(msg_resp['fields'])):
                            field = msg_resp['fields'][i]
                            if field['title'] == 'Last Updated':
                                new_fields[i]['value'] = str(time.strftime("%Y-%m-%d %H:%M:%S"))
                        msg_resp['fields'] = new_fields
                    message_id = msg['ts']
                    self.client.chat_update(channel=channel_id,ts=message_id,text="",attachments=[msg_resp])

    def send_alert_slack(self, workcell:str, tool:str, protocol:str, error:str, channel_id:str) -> None:
        message = self.create_alert_message(title="Workcell has errored!", workcell=workcell, tool=tool, protocol=protocol, error=error)
        self.slack_block_message(message="Galago Alert",attachments=message,recipient=channel_id )
    
    def retrieve_message(self, message_id:str, channel_id:str)-> Any:
        try:
            result = self.client.conversations_history(
                channel=channel_id,
                inclusive=True,
                oldest=message_id,
                limit=1
            )
            print("Raw message is"+ str(result))
            return result
        except Exception as e:
            logging.error(f"Error: {e}")

    def get_last_messages(self, channel_id:str) -> Any:
        try:
            response = self.client.conversations_history(channel=channel_id, limit=5)
            if response['ok'] and response['messages']:
                last_messages = response['messages']
                #print(f"Last 5 messages {last_messages}")
                return last_messages
            else:
                return None
        except Exception as e:
            logging.error(f"Error fetching messages from Slack: {e}")
            return None




import os 
import subprocess 
import logging 

def check_conda_is_path() -> bool:
    if os.name != 'nt':
        raise RuntimeError("This method is meant to be used on a windows machine only")
    cmd = "conda"
    try:
        process = subprocess.Popen(cmd,stdout=subprocess.PIPE)
        process.communicate()[0]
        return True
    except Exception as e:
        logging.error(f"Failed to call conda command. {e}")
        return False

def conda_activate(env_name:str) -> bool:
    if os.name != 'nt':
        raise RuntimeError("This method is meant to be used on a windows machine only")

    cmd = ["conda","activate",env_name]
    try:
        process = subprocess.Popen(cmd,stdout=subprocess.PIPE,shell=True)
        response = process.communicate()[0]
        logging.info(response)
        return True
    except Exception as e:
        logging.error(f"Failed to call conda command. {e}")
        return False
    
def get_conda_environments() -> list[str]:
    if os.name != 'nt':
        raise RuntimeError("This method is meant to be used on a windows machine only")
    conda_is_path = check_conda_is_path()
    paths = []
    if conda_is_path:
        cmd = ["conda","info","--envs"]
        process = subprocess.Popen(cmd,stdout=subprocess.PIPE)
        response = process.communicate()[0]
        decoded_string = response.decode('utf-8')
        lines = decoded_string.split('\r\n')
        for line in lines:
            splitted_line = line.split()
            if len(splitted_line) > 1:
                if splitted_line[0] == "#":
                    continue
                env_name = splitted_line[0]
                paths.append(env_name)

        return paths
    else:
        return paths



import typing as t
import logging
import os
import re
import tempfile
import time

from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.opentrons2_pb2 import Command, Config
from google.protobuf import json_format
from tools.app_config import Config as AppConfig 
from .driver import Ot2Driver
import argparse

class Opentrons2Server(ToolServer):
    toolType = "opentrons2"
    driver: Ot2Driver
    config: Config
    
    def __init__(self) -> None:
          super().__init__()
          self.app_config = AppConfig()
          self.driver_config : Config 
          self.process_images : bool = True 
          
    def _configure(self, config: Config) -> None:
        self.driver_config = config
        self.driver = Ot2Driver(robot_ip=config.robot_ip, robot_port=config.robot_port)
        self.driver.ping()

    def _createProgramFromTemplate(self, templatePath: str, params: dict[t.Any, t.Any]) -> str:
        """
        Creates a program from a template and parameters, creates a new temporary file and returns the path to it
        """
        template = open(templatePath).read()
        params_str = str(params)
        # HACK: strip out the trailing .0 on any numbers, because protobuf turns all ints into floats
        params_str = re.sub(r"(\d)\.0\b", r"\1", params_str)
        program = re.sub(
            r"#.*PARAMS_START.*\n[\s\S]*\n#.*PARAMS_END.*\n",
            "params = %s\n" % params_str,
            template,
            re.MULTILINE,
        )
        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".py") as f:
            f.write(program)
            logging.info(f"Created temporary program {f.name}")
            return f.name

    def RunProgram(self, params: Command.RunProgram) -> None:
        #today : str = datetime.today().strftime('%Y-%m-%d')
        #pre_run_name = f"ot2_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}_pre_run.jpg"
        #self.driver.take_picture(pre_run_name, os.path.join(self.app_config.app_config.data_folder,"images","ot2",today))
        if params.program_name == "universal_opentrons_json_executor":
            program_dir = os.path.dirname(os.path.abspath(__file__))
        else:
            program_dir = self.config.program_dir
        program_path = os.path.join(program_dir, params.program_name + ".py")
        program_params = params.params
        logging.info(f"Running program {program_path} with params {program_params}")
        params_dict = json_format.MessageToDict(program_params)
        program = self._createProgramFromTemplate(templatePath=program_path, params=params_dict)
        self.driver.start_protocol(protocol_file=program)
       # post_run_name = f"ot2_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}_pre_run.jpg"
        #self.driver.take_picture(post_run_name, os.path.join(self.app_config.app_config.data_folder,"images","ot2",today))
    
    def Sleep(self, params: Command.Sleep) -> None:
        logging.info(f"Sleeping for {params.seconds} seconds")
        time.sleep(params.seconds)

    def Pause(self, params: Command.Pause) -> None:
        logging.info("Pausing program")
        self.driver.pause_protocol()

    def Resume(self, params: Command.Resume) -> None:
        logging.info("Resuming program")
        self.driver.resume_protocol()

    def Cancel(self, params: Command.Cancel) -> None:
        logging.info("Canceling program")
        self.driver.cancel_protocol()

    def ToggleLight(self, params: Command.ToggleLight) -> None:
        logging.info("Toggling light")
        self.driver.toggle_light()

    def EstimateRunProgram(self, params: Command.RunProgram) -> int:
        return 1

    def EstimateSleep(self, params: Command.Sleep) -> int:
        return 1

    def EstimatePause(self, params: Command.Pause) -> int:
        return 1

    def EstimateResume(self, params: Command.Resume) -> int:
        return 1

    def EstimateCancel(self, params: Command.Cancel) -> int:
        return 1

    def EstimateToggleLight(self, params: Command.ToggleLight) -> int:
        return 1


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
         raise RuntimeWarning("Port must be provided...")
    serve(Opentrons2Server(), os.environ.get("PORT", str(args.port)))



import logging
import os

from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.vcode_pb2 import Command, Config
from .driver import VCodeDriver
import argparse 

class VCodeServer(ToolServer):
    toolType = "vcode"

    def __init__(self) -> None:
        super().__init__()
        self.driver : VCodeDriver

    def _configure(self, request:Config) -> None:
        logging.info("Initializing VCode...")
        self.config = request
        self.driver = VCodeDriver(request.profile)
        return self.driver.initialize()
    
    def Home(self, params:Command.Home) -> None:
        return self.driver.home_stage()
    
    def PrintAndApply(self, params:Command.PrintAndApply) -> None:
        return self.driver.print_and_apply_by_name(
            params.format_name,
            params.side, 
            params.drop_stage, 
            params.field_0,
            params.field_1,
            params.field_2,
            params.field_3,
            params.field_4,
            params.field_5
        )
    
    def Print(self, params:Command.Print) -> None:
        return self.driver.print_label(
            params.format_name,
            params.field_0,
            params.field_1,
            params.field_2,
            params.field_3,
            params.field_4,
            params.field_5
        )
    
    def Rotate180(self, params:Command.Rotate180) -> None:
        return self.driver.rotate180()
    
    def RotateStage(self, params:Command.RotateStage) -> None:
        return self.driver.rotate_stage(params.angle)
    
    def ShowDiagsDialog(self, params:Command.ShowDiagsDialog) -> None:
        return self.driver.show_diagnostics()

    def DropStage(self, params:Command.DropStage) -> None:
        return self.driver.drop_stage(params.drop_stage)
    
    def EstimatePrintAndApply(self, params: Command.PrintAndApply) -> int:
        return 1
    
    def EstimateHome(self, params: Command.Home) -> int:
        return 1
    
    def EstimatePrint(self, params: Command.Print) -> int:
        return 1
    
    def EstimateRotate180(self, params: Command.Rotate180) -> int:
        return 1
    
    def EstimateDropStage(self, params: Command.DropStage) -> int:
        return 1
    
    def EstimateShowDiagsDialog(self, params: Command.ShowDiagsDialog) -> int:
        return 1
    
    def EstimateRotateStage(self, params: Command.RotateStage) -> int:
        return 1

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
         raise RuntimeWarning("Port must be provided...")
    serve(VCodeServer(), os.environ.get("PORT",str(args.port)))



# -*- coding: utf-8 -*-
# Created by makepy.py version 0.5.01
# By python version 3.10.4 | packaged by conda-forge | (main, Mar 30 2022, 08:38:02) [MSC v.1916 64 bit (AMD64)]
# From type library 'Gen5.exe'
# On Thu Aug 17 16:26:05 2023
''
makepy_version = '0.5.01'
python_version = 0x30a04f0

import win32com.client.CLSIDToClass, pythoncom, pywintypes
import win32com.client.util
from pywintypes import IID
from win32com.client import Dispatch

# The following 3 lines may need tweaking for the particular server
# Candidates are pythoncom.Missing, .Empty and .ArgNotFound
defaultNamedOptArg=pythoncom.Empty
defaultNamedNotOptArg=pythoncom.Empty
defaultUnnamedArg=pythoncom.Empty

CLSID = IID('{80BC3D0F-EA45-49B3-8F79-4E0B1D635711}')
MajorVersion = 1
MinorVersion = 0
LibraryFlags = 8
LCID = 0x0

class constants:
	eCalculationCompleted         =3          # from enum Gen5CalculationStatus
	eCalculationNotStarted        =0          # from enum Gen5CalculationStatus
	eCalculationPaused            =2          # from enum Gen5CalculationStatus
	eCalculationRunning           =1          # from enum Gen5CalculationStatus
	eCalculationStatusUnknown     =-1         # from enum Gen5CalculationStatus
	eOLECommunication             =1          # from enum Gen5CommunicationType
	eSerialCommunication          =0          # from enum Gen5CommunicationType
	eFileTypeExp                  =1          # from enum Gen5FileType
	eFileTypeOpn                  =4          # from enum Gen5FileType
	eFileTypePnl                  =8          # from enum Gen5FileType
	eFileTypePrt                  =2          # from enum Gen5FileType
	eFileTypeXts                  =16         # from enum Gen5FileType
	eRawDataFoundWithMore         =1          # from enum Gen5GetRawDataStatus
	eRawDataFoundWithNoMore       =2          # from enum Gen5GetRawDataStatus
	eRawDataNoneAvailable         =0          # from enum Gen5GetRawDataStatus
	eFromProtocolFile             =0          # from enum Gen5NewExperimentMode
	eFromProtocolMetadata         =1          # from enum Gen5NewExperimentMode
	ePlateTypeCalibration         =1          # from enum Gen5PlateType
	ePlateTypeStandard            =0          # from enum Gen5PlateType
	ePlateTypeSelectionFavorite   =1          # from enum Gen5PlateTypeSelection
	ePlateTypeSelectionHasLid     =8          # from enum Gen5PlateTypeSelection
	ePlateTypeSelectionTypeCustom =4          # from enum Gen5PlateTypeSelection
	ePlateTypeSelectionTypeDefault=2          # from enum Gen5PlateTypeSelection
	eWellStatusError              =4          # from enum Gen5PrimaryWellStatus
	eWellStatusMasked             =3          # from enum Gen5PrimaryWellStatus
	eWellStatusMissedValue        =2          # from enum Gen5PrimaryWellStatus
	eWellStatusNoValue            =5          # from enum Gen5PrimaryWellStatus
	eWellStatusOK                 =0          # from enum Gen5PrimaryWellStatus
	eWellStatusOutOfRange         =1          # from enum Gen5PrimaryWellStatus
	eProtocolTypeCalibration      =1          # from enum Gen5ProtocolType
	eProtocolTypeMultiPlateAssay  =2          # from enum Gen5ProtocolType
	eProtocolTypeStandard         =0          # from enum Gen5ProtocolType
	eReadAborted                  =2          # from enum Gen5ReadStatus
	eReadCompleted                =5          # from enum Gen5ReadStatus
	eReadError                    =4          # from enum Gen5ReadStatus
	eReadInProgress               =1          # from enum Gen5ReadStatus
	eReadNotStarted               =0          # from enum Gen5ReadStatus
	eReadPaused                   =3          # from enum Gen5ReadStatus
	eAbsorbanceReadModeCount      =3          # from enum Gen5ReaderCharacteristicsID
	eAbsorbanceReadModeName       =4          # from enum Gen5ReaderCharacteristicsID
	eAbsorbanceReadSuported       =0          # from enum Gen5ReaderCharacteristicsID
	eAbsorbanceWavelengthMax      =2          # from enum Gen5ReaderCharacteristicsID
	eAbsorbanceWavelengthMin      =1          # from enum Gen5ReaderCharacteristicsID
	eFilterFluorescenceSupported  =12         # from enum Gen5ReaderCharacteristicsID
	eInstrumentName               =11         # from enum Gen5ReaderCharacteristicsID
	eMonoFluorescenceSupported    =13         # from enum Gen5ReaderCharacteristicsID
	eReaderArchitectureLevel      =14         # from enum Gen5ReaderCharacteristicsID
	eSerialNumber                 =10         # from enum Gen5ReaderCharacteristicsID
	eShakeSupported               =9          # from enum Gen5ReaderCharacteristicsID
	eTemperatureControlOption     =5          # from enum Gen5ReaderCharacteristicsID
	eTemperatureGradientMax       =8          # from enum Gen5ReaderCharacteristicsID
	eTemperatureMax               =7          # from enum Gen5ReaderCharacteristicsID
	eTemperatureMin               =6          # from enum Gen5ReaderCharacteristicsID
	eReaderCommand_CleanObjective =2048       # from enum Gen5ReaderControlCommand
	eReaderCommand_ControlDispenser=8          # from enum Gen5ReaderControlCommand
	eReaderCommand_ControlDoor    =2          # from enum Gen5ReaderControlCommand
	eReaderCommand_ControlGasCharge=128        # from enum Gen5ReaderControlCommand
	eReaderCommand_ControlIncubator=16         # from enum Gen5ReaderControlCommand
	eReaderCommand_ControlLamp    =32         # from enum Gen5ReaderControlCommand
	eReaderCommand_ControlPrime   =64         # from enum Gen5ReaderControlCommand
	eReaderCommand_ControlShake   =4          # from enum Gen5ReaderControlCommand
	eReaderCommand_DoorIn         =512        # from enum Gen5ReaderControlCommand
	eReaderCommand_DoorOut        =1024       # from enum Gen5ReaderControlCommand
	eReaderCommand_ShowControlPanel=0          # from enum Gen5ReaderControlCommand
	eReaderCommand_ShowInformation=1          # from enum Gen5ReaderControlCommand
	eReaderStatus_Busy            =-1         # from enum Gen5ReaderStatus
	eReaderStatus_NotCommunicating=-2         # from enum Gen5ReaderStatus
	eReaderStatus_NotConfigured   =-3         # from enum Gen5ReaderStatus
	eReaderStatus_OK              =0          # from enum Gen5ReaderStatus
	eSecondaryStatusInjector1     =1          # from enum Gen5SecondaryWellStatus
	eSecondaryStatusInjector2     =2          # from enum Gen5SecondaryWellStatus
	eSecondaryStatusInjector3     =3          # from enum Gen5SecondaryWellStatus
	eSecondaryStatusInjector4     =4          # from enum Gen5SecondaryWellStatus
	eSecondaryStatusOK            =0          # from enum Gen5SecondaryWellStatus
	eTemperatureDirectFromReader  =1          # from enum Gen5TemperatureStatus
	eTemperatureInValid           =-2         # from enum Gen5TemperatureStatus
	eTemperatureNotInitialized    =-1         # from enum Gen5TemperatureStatus
	eTemperatureNotSupported      =-3         # from enum Gen5TemperatureStatus
	eTemperatureObtainedWithReaderData=2          # from enum Gen5TemperatureStatus

from win32com.client import DispatchBaseClass
class IApplication(DispatchBaseClass):
	CLSID = IID('{56BBF0D8-DDE6-4F25-BEE1-F3A7CC516531}')
	coclass_clsid = IID('{9B3B05F7-143C-4DAF-93D8-726A7BA3321A}')

	def BrowseForFile(self, OpenFileDlg=defaultNamedNotOptArg, FileName=defaultNamedNotOptArg, DialogTitle=defaultNamedNotOptArg, eFileType=defaultNamedNotOptArg):
		'method BrowseForFile'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(11, LCID, 1, (8, 0), ((11, 0), (8, 0), (8, 0), (3, 0)),OpenFileDlg
			, FileName, DialogTitle, eFileType)

	def BrowseForFolder(self, FolderName=defaultNamedNotOptArg):
		'method BrowseForFolder'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(12, LCID, 1, (8, 0), ((8, 0),),FolderName
			)

	def CarrierIn(self):
		'method CarrierIn'
		return self._oleobj_.InvokeTypes(13, LCID, 1, (24, 0), (),)

	def CarrierInEx(self, p_pszPlateTypeName=defaultNamedNotOptArg, p_bLidOnPlate=defaultNamedNotOptArg):
		'method CarrierInEx'
		return self._oleobj_.InvokeTypes(37, LCID, 1, (24, 0), ((8, 0), (11, 0)),p_pszPlateTypeName
			, p_bLidOnPlate)

	def CarrierOut(self):
		'method CarrierOut'
		return self._oleobj_.InvokeTypes(14, LCID, 1, (24, 0), (),)

	def ClearTipPrimeTrough(self):
		'method ClearTipPrimeTrough'
		return self._oleobj_.InvokeTypes(25, LCID, 1, (24, 0), (),)

	def ConfigureOleReader(self, ReaderType=defaultNamedNotOptArg):
		'method ConfigureOleReader'
		return self._oleobj_.InvokeTypes(8, LCID, 1, (24, 0), ((3, 0),),ReaderType
			)

	def ConfigureSerialReader(self, ReaderType=defaultNamedNotOptArg, ComPort=defaultNamedNotOptArg, BaudRate=defaultNamedNotOptArg):
		'method ConfigureSerialReader'
		return self._oleobj_.InvokeTypes(9, LCID, 1, (24, 0), ((3, 0), (3, 0), (3, 0)),ReaderType
			, ComPort, BaudRate)

	def ConfigureUSBReader(self, ReaderType=defaultNamedNotOptArg, ReaderSN=defaultNamedNotOptArg):
		'method ConfigureUSBReader'
		return self._oleobj_.InvokeTypes(23, LCID, 1, (24, 0), ((3, 0), (8, 0)),ReaderType
			, ReaderSN)

	def DiagTestPlateRunDlg(self):
		'method DiagTestPlateRunDlg'
		return self._oleobj_.InvokeTypes(45, LCID, 1, (24, 0), (),)

	def DisplayAbsorbanceTestPlateDlg(self):
		'method DisplayAbsorbanceTestPlateDlg'
		return self._oleobj_.InvokeTypes(44, LCID, 1, (24, 0), (),)

	def DisplayImagingLiveModeDialog(self):
		'method DisplayImagingLiveModeDialog'
		return self._oleobj_.InvokeTypes(41, LCID, 1, (24, 0), (),)

	def EditPlateTypes(self):
		'method EditPlateTypes'
		return self._oleobj_.InvokeTypes(29, LCID, 1, (24, 0), (),)

	def GetBioStackInterfaceID(self, ReaderType=defaultNamedNotOptArg):
		'method GetBioStackInterfaceID'
		return self._oleobj_.InvokeTypes(20, LCID, 1, (3, 0), ((3, 0),),ReaderType
			)

	def GetConfiguredInstruments(self):
		'method GetConfiguredInstruments'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(39, LCID, 1, (8, 0), (),)

	def GetCurrentTemperature(self, pvTemperatureValue=defaultNamedNotOptArg, pnTemperatureStatus=defaultNamedNotOptArg):
		'method GetCurrentTemperature'
		return self._oleobj_.InvokeTypes(26, LCID, 1, (24, 0), ((16396, 0), (16396, 0)),pvTemperatureValue
			, pnTemperatureStatus)

	def GetCurrentTemperatureFP(self, pvTemperatureValue=defaultNamedNotOptArg, pnTemperatureStatus=defaultNamedNotOptArg):
		'method GetCurrentTemperatureFP'
		return self._oleobj_.InvokeTypes(42, LCID, 1, (24, 0), ((16396, 0), (16396, 0)),pvTemperatureValue
			, pnTemperatureStatus)

	def GetHotelInterfaceID(self, ReaderType=defaultNamedNotOptArg):
		'method GetHotelInterfaceID'
		return self._oleobj_.InvokeTypes(30, LCID, 1, (3, 0), ((3, 0),),ReaderType
			)

	def GetLastReaderError(self):
		'method GetLastReaderError'
		return self._oleobj_.InvokeTypes(15, LCID, 1, (3, 0), (),)

	def GetMonoCalTestResults(self, pvPeakWavelength=defaultNamedNotOptArg, pvError=defaultNamedNotOptArg):
		'method GetMonoCalTestResults'
		return self._oleobj_.InvokeTypes(22, LCID, 1, (3, 0), ((16396, 0), (16396, 0)),pvPeakWavelength
			, pvError)

	def GetPlateTypeDefinition(self, p_strName=defaultNamedNotOptArg):
		'method GetPlateTypeDefinition'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(27, LCID, 1, (8, 0), ((8, 0),),p_strName
			)

	def GetPlateTypeNames(self, p_nSelectionFlag=defaultNamedNotOptArg, pvNames=defaultNamedNotOptArg):
		'method GetPlateTypeNames'
		return self._oleobj_.InvokeTypes(28, LCID, 1, (24, 0), ((2, 0), (16396, 0)),p_nSelectionFlag
			, pvNames)

	def GetReaderCharacteristics(self, ReaderCharacteristicsID=defaultNamedNotOptArg, ReaderCharacteristicsIndex=defaultNamedNotOptArg, pvValue=defaultNamedNotOptArg):
		'method GetReaderCharacteristics'
		return self._oleobj_.InvokeTypes(35, LCID, 1, (24, 0), ((3, 0), (3, 0), (16396, 0)),ReaderCharacteristicsID
			, ReaderCharacteristicsIndex, pvValue)

	def GetReaderStatus(self):
		'method GetReaderStatus'
		return self._oleobj_.InvokeTypes(34, LCID, 1, (3, 0), (),)

	def GetSupportedReaders(self, pvReaderName=defaultNamedNotOptArg, pvReaderType=defaultNamedNotOptArg, pvCommunicationType=defaultNamedNotOptArg):
		'method GetSupportedReaders'
		return self._oleobj_.InvokeTypes(19, LCID, 1, (24, 0), ((16396, 0), (16396, 0), (16396, 0)),pvReaderName
			, pvReaderType, pvCommunicationType)

	def GetTemperatureSetPoint(self, pvSetpoint=defaultNamedNotOptArg, pvGradient=defaultNamedNotOptArg):
		'method GetTemperatureSetPoint'
		return self._oleobj_.InvokeTypes(36, LCID, 1, (24, 0), ((16396, 0), (16396, 0)),pvSetpoint
			, pvGradient)

	def ImportPlateTypeDefinition(self, p_strImportPlateTypeXML=defaultNamedNotOptArg):
		'method ImportPlateTypeDefinition'
		return self._oleobj_.InvokeTypes(40, LCID, 1, (24, 0), ((8, 0),),p_strImportPlateTypeXML
			)

	def NewExperiment(self, ProtocolPathName=defaultNamedNotOptArg):
		'method NewExperiment'
		ret = self._oleobj_.InvokeTypes(6, LCID, 1, (9, 0), ((8, 0),),ProtocolPathName
			)
		if ret is not None:
			ret = Dispatch(ret, 'NewExperiment', None)
		return ret

	def NewExperimentEx(self, CreationMode=defaultNamedNotOptArg, ProtocolMetadata=defaultNamedNotOptArg):
		'method NewExperimentEx'
		ret = self._oleobj_.InvokeTypes(33, LCID, 1, (9, 0), ((3, 0), (8, 0)),CreationMode
			, ProtocolMetadata)
		if ret is not None:
			ret = Dispatch(ret, 'NewExperimentEx', None)
		return ret

	def OpenClinicalDatabaseGroup(self, DatabasePathname=defaultNamedNotOptArg, Group=defaultNamedNotOptArg, UseQCRTool=defaultNamedNotOptArg):
		'method OpenClinicalDatabaseGroup'
		return self._oleobj_.InvokeTypes(17, LCID, 1, (24, 0), ((8, 0), (8, 0), (11, 0)),DatabasePathname
			, Group, UseQCRTool)

	def OpenExperiment(self, PathName=defaultNamedNotOptArg):
		'method OpenExperiment'
		ret = self._oleobj_.InvokeTypes(7, LCID, 1, (9, 0), ((8, 0),),PathName
			)
		if ret is not None:
			ret = Dispatch(ret, 'OpenExperiment', None)
		return ret

	def RunReaderControlCommand(self, p_nCommand=defaultNamedNotOptArg):
		'method RunReaderControlCommand'
		return self._oleobj_.InvokeTypes(31, LCID, 1, (24, 0), ((3, 0),),p_nCommand
			)

	def SetClientWindow(self, ClientWindowHandle=defaultNamedNotOptArg):
		'method SetClientWindow'
		return self._oleobj_.InvokeTypes(10, LCID, 1, (24, 0), ((3, 0),),ClientWindowHandle
			)

	def SetTemperatureSetPoint(self, p_bIncubatorState=defaultNamedNotOptArg, p_nTemperatureSetPoint=defaultNamedNotOptArg, p_nGradient=defaultNamedNotOptArg):
		'method SetTemperatureSetPoint'
		return self._oleobj_.InvokeTypes(32, LCID, 1, (24, 0), ((11, 0), (2, 0), (2, 0)),p_bIncubatorState
			, p_nTemperatureSetPoint, p_nGradient)

	def SimpleRequestResponseHandler(self, p_chCommand=defaultNamedNotOptArg, p_byCommandBody=defaultNamedNotOptArg, p_nCommandBodyLength=defaultNamedNotOptArg, p_pbyResponseStream=defaultNamedNotOptArg
			, p_nExpectedResponseLength=defaultNamedNotOptArg, p_lMSecTimeoutValue=defaultNamedNotOptArg, p_plReaderHexError=defaultNamedNotOptArg):
		'method SimpleRequestResponseHandler'
		return self._oleobj_.InvokeTypes(43, LCID, 1, (2, 0), ((8, 0), (8, 0), (2, 0), (16392, 0), (2, 0), (3, 0), (16387, 0)),p_chCommand
			, p_byCommandBody, p_nCommandBodyLength, p_pbyResponseStream, p_nExpectedResponseLength, p_lMSecTimeoutValue
			, p_plReaderHexError)

	def StartMonoCalTest(self, PathName=defaultNamedNotOptArg):
		'method StartMonoCalTest'
		ret = self._oleobj_.InvokeTypes(21, LCID, 1, (9, 0), ((8, 0),),PathName
			)
		if ret is not None:
			ret = Dispatch(ret, 'StartMonoCalTest', None)
		return ret

	def StartSystemTest(self, p_bForceNewSystemTest=defaultNamedNotOptArg):
		'method StartSystemTest'
		ret = self._oleobj_.InvokeTypes(38, LCID, 1, (9, 0), ((11, 0),),p_bForceNewSystemTest
			)
		if ret is not None:
			ret = Dispatch(ret, 'StartSystemTest', None)
		return ret

	def TestReaderCommunication(self):
		'method TestReaderCommunication'
		return self._oleobj_.InvokeTypes(16, LCID, 1, (3, 0), (),)

	_prop_map_get_ = {
		"AllowInterruptedDiscontinuousKineticResume": (46, 2, (11, 0), (), "AllowInterruptedDiscontinuousKineticResume", None),
		"CreateTempFileIfUnsavedExperiment": (24, 2, (11, 0), (), "CreateTempFileIfUnsavedExperiment", None),
		"DataExportEnabled": (18, 2, (11, 0), (), "DataExportEnabled", None),
		"DatabaseFileStorage": (5, 2, (11, 0), (), "DatabaseFileStorage", None),
		"Gen5AppPath": (4, 2, (8, 0), (), "Gen5AppPath", None),
		"Gen5BuildNumber": (3, 2, (3, 0), (), "Gen5BuildNumber", None),
		"Gen5VersionNumber": (2, 2, (5, 0), (), "Gen5VersionNumber", None),
		"Gen5VersionString": (1, 2, (8, 0), (), "Gen5VersionString", None),
	}
	_prop_map_put_ = {
		"AllowInterruptedDiscontinuousKineticResume" : ((46, LCID, 4, 0),()),
		"CreateTempFileIfUnsavedExperiment" : ((24, LCID, 4, 0),()),
		"DataExportEnabled" : ((18, LCID, 4, 0),()),
		"DatabaseFileStorage" : ((5, LCID, 4, 0),()),
		"Gen5AppPath" : ((4, LCID, 4, 0),()),
		"Gen5BuildNumber" : ((3, LCID, 4, 0),()),
		"Gen5VersionNumber" : ((2, LCID, 4, 0),()),
		"Gen5VersionString" : ((1, LCID, 4, 0),()),
	}
	def __iter__(self):
		"Return a Python iterator for this object"
		try:
			ob = self._oleobj_.InvokeTypes(-4,LCID,3,(13, 10),())
		except pythoncom.error:
			raise TypeError("This object does not support enumeration")
		return win32com.client.util.Iterator(ob, None)

class IExperiment(DispatchBaseClass):
	CLSID = IID('{21258B83-0AAE-43F7-AFBC-89C680D51C6F}')
	coclass_clsid = IID('{F4024CA7-7362-4FC8-B33F-F1F2BD58788F}')

	def AddAuditTrailRecord(self, p_pszDescription=defaultNamedNotOptArg, p_pszComment=defaultNamedNotOptArg):
		'method AddAuditTrailRecord'
		return self._oleobj_.InvokeTypes(15, LCID, 1, (24, 0), ((8, 0), (8, 0)),p_pszDescription
			, p_pszComment)

	def Close(self):
		'method Close'
		return self._oleobj_.InvokeTypes(8, LCID, 1, (24, 0), (),)

	def GetPlateLayout(self, ProtocolID=defaultNamedNotOptArg):
		'method GetPlateLayout'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(10, LCID, 1, (8, 0), ((8, 0),),ProtocolID
			)

	def GetSampleIdentification(self):
		'method GetSampleIdentification'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(12, LCID, 1, (8, 0), (),)

	def Save(self):
		'method Save'
		return self._oleobj_.InvokeTypes(7, LCID, 1, (24, 0), (),)

	def SaveAs(self, PathName=defaultNamedNotOptArg):
		'method SaveAs'
		return self._oleobj_.InvokeTypes(9, LCID, 1, (24, 0), ((8, 0),),PathName
			)

	def SetPlateLayout(self, PlateLayoutXML=defaultNamedNotOptArg):
		'method SetPlateLayout'
		return self._oleobj_.InvokeTypes(11, LCID, 1, (24, 0), ((8, 0),),PlateLayoutXML
			)

	def SetSampleIdentification(self, SampleIdentificationXML=defaultNamedNotOptArg):
		'method SetSampleIdentification'
		return self._oleobj_.InvokeTypes(13, LCID, 1, (24, 0), ((8, 0),),SampleIdentificationXML
			)

	_prop_map_get_ = {
		"AutoFileExport": (4, 2, (11, 0), (), "AutoFileExport", None),
		"AutoPowerExport": (5, 2, (11, 0), (), "AutoPowerExport", None),
		"AutoPrintReport": (3, 2, (11, 0), (), "AutoPrintReport", None),
		"AutoSave": (2, 2, (11, 0), (), "AutoSave", None),
		"DeleteObsoleteDRFoldersOnSave": (14, 2, (11, 0), (), "DeleteObsoleteDRFoldersOnSave", None),
		"Plates": (1, 2, (9, 0), (), "Plates", None),
		"ProtocolType": (6, 2, (3, 0), (), "ProtocolType", None),
	}
	_prop_map_put_ = {
		"AutoFileExport" : ((4, LCID, 4, 0),()),
		"AutoPowerExport" : ((5, LCID, 4, 0),()),
		"AutoPrintReport" : ((3, LCID, 4, 0),()),
		"AutoSave" : ((2, LCID, 4, 0),()),
		"DeleteObsoleteDRFoldersOnSave" : ((14, LCID, 4, 0),()),
		"Plates" : ((1, LCID, 4, 0),()),
		"ProtocolType" : ((6, LCID, 4, 0),()),
	}
	def __iter__(self):
		"Return a Python iterator for this object"
		try:
			ob = self._oleobj_.InvokeTypes(-4,LCID,3,(13, 10),())
		except pythoncom.error:
			raise TypeError("This object does not support enumeration")
		return win32com.client.util.Iterator(ob, None)

class IMonoCalTestMonitor(DispatchBaseClass):
	CLSID = IID('{0DABEE12-4422-4FE2-8D4B-F510CA25ECD7}')
	coclass_clsid = IID('{9EFA02E9-6776-44B3-B193-62A360356658}')

	def GetErrorMessage(self):
		'method GetErrorMessage'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(3, LCID, 1, (8, 0), (),)

	_prop_map_get_ = {
		"TestInProgress": (1, 2, (11, 0), (), "TestInProgress", None),
		"TestSuccessful": (2, 2, (11, 0), (), "TestSuccessful", None),
	}
	_prop_map_put_ = {
		"TestInProgress" : ((1, LCID, 4, 0),()),
		"TestSuccessful" : ((2, LCID, 4, 0),()),
	}
	def __iter__(self):
		"Return a Python iterator for this object"
		try:
			ob = self._oleobj_.InvokeTypes(-4,LCID,3,(13, 10),())
		except pythoncom.error:
			raise TypeError("This object does not support enumeration")
		return win32com.client.util.Iterator(ob, None)

class IPlate(DispatchBaseClass):
	CLSID = IID('{450BF0EA-F196-4D18-AD0D-DB39D2E6E08B}')
	coclass_clsid = IID('{DA2A9CC3-FE47-441B-BD0C-B482015A525F}')

	def AbortRead(self):
		'method AbortRead'
		return self._oleobj_.InvokeTypes(11, LCID, 1, (24, 0), (),)

	def AddAuditTrailRecord(self, p_pszDescription=defaultNamedNotOptArg, p_pszComment=defaultNamedNotOptArg):
		'method AddAuditTrailRecord'
		return self._oleobj_.InvokeTypes(69, LCID, 1, (24, 0), ((8, 0), (8, 0)),p_pszDescription
			, p_pszComment)

	def AsyncMovieExport(self, p_pszBuilderName=defaultNamedNotOptArg, p_pszFolderName=defaultNamedNotOptArg):
		'method AsyncMovieExport'
		ret = self._oleobj_.InvokeTypes(66, LCID, 1, (9, 0), ((8, 0), (8, 0)),p_pszBuilderName
			, p_pszFolderName)
		if ret is not None:
			ret = Dispatch(ret, 'AsyncMovieExport', None)
		return ret

	def AsyncPictureExport(self, p_pszBuilderName=defaultNamedNotOptArg, p_pszFolderName=defaultNamedNotOptArg):
		'method AsyncPictureExport'
		ret = self._oleobj_.InvokeTypes(65, LCID, 1, (9, 0), ((8, 0), (8, 0)),p_pszBuilderName
			, p_pszFolderName)
		if ret is not None:
			ret = Dispatch(ret, 'AsyncPictureExport', None)
		return ret

	def Delete(self):
		'method Delete'
		return self._oleobj_.InvokeTypes(14, LCID, 1, (24, 0), (),)

	def DisableKeepReadProcessGoingUntilCalculationsCompleted(self):
		'method DisableKeepReadProcessGoingUntilCalculationsCompleted'
		return self._oleobj_.InvokeTypes(55, LCID, 1, (24, 0), (),)

	def FileExport(self, ExportFilePath=defaultNamedNotOptArg):
		'method FileExport'
		return self._oleobj_.InvokeTypes(13, LCID, 1, (24, 0), ((8, 0),),ExportFilePath
			)

	def FileExportEx(self, p_pszBuilderName=defaultNamedNotOptArg, p_pszfilePath=defaultNamedNotOptArg):
		'method FileExportEx'
		return self._oleobj_.InvokeTypes(32, LCID, 1, (24, 0), ((8, 0), (8, 0)),p_pszBuilderName
			, p_pszfilePath)

	def GetChannelBCValues(self, ChannelName=defaultNamedNotOptArg, pBrightnessValue=defaultNamedNotOptArg, p_ContrastValue=defaultNamedNotOptArg):
		'method GetChannelBCValues'
		return self._oleobj_.InvokeTypes(61, LCID, 1, (24, 0), ((8, 0), (16396, 0), (16396, 0)),ChannelName
			, pBrightnessValue, p_ContrastValue)

	def GetDataReductionInfo(self, ProtocolID=defaultNamedNotOptArg):
		'method GetDataReductionInfo'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(58, LCID, 1, (8, 0), ((8, 0),),ProtocolID
			)

	def GetDataSetInfo(self, pvDataSetName=defaultNamedNotOptArg, pvDecimalPlaces=defaultNamedNotOptArg, pvKineticReads=defaultNamedNotOptArg, pvKineticIntervalMSec=defaultNamedNotOptArg
			, pvDisplayMSec=defaultNamedNotOptArg, pvSpectrumWavelengths=defaultNamedNotOptArg, pvSpectrumStart=defaultNamedNotOptArg, pvSpectrumStep=defaultNamedNotOptArg, pvHorizontalReads=defaultNamedNotOptArg
			, pvVerticalReads=defaultNamedNotOptArg):
		'method GetDataSetInfo'
		return self._oleobj_.InvokeTypes(24, LCID, 1, (24, 0), ((16396, 0), (16396, 0), (16396, 0), (16396, 0), (16396, 0), (16396, 0), (16396, 0), (16396, 0), (16396, 0), (16396, 0)),pvDataSetName
			, pvDecimalPlaces, pvKineticReads, pvKineticIntervalMSec, pvDisplayMSec, pvSpectrumWavelengths
			, pvSpectrumStart, pvSpectrumStep, pvHorizontalReads, pvVerticalReads)

	def GetDataSetInfoEx(self, p_pszProtocolID=defaultNamedNotOptArg, p_pszDataSetName=defaultNamedNotOptArg):
		'method GetDataSetInfoEx'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(49, LCID, 1, (8, 0), ((8, 0), (8, 0)),p_pszProtocolID
			, p_pszDataSetName)

	def GetDataSetNames(self, p_pszProtocolID=defaultNamedNotOptArg, pvNames=defaultNamedNotOptArg):
		'method GetDataSetNames'
		return self._oleobj_.InvokeTypes(48, LCID, 1, (24, 0), ((8, 0), (16396, 0)),p_pszProtocolID
			, pvNames)

	def GetDataSetROIsInWell(self, p_pszProtocolID=defaultNamedNotOptArg, p_pszDataSetName=defaultNamedNotOptArg, p_pszWellIndexes=defaultNamedNotOptArg):
		'method GetDataSetROIsInWell'
		return self._oleobj_.InvokeTypes(64, LCID, 1, (3, 0), ((8, 0), (8, 0), (8, 0)),p_pszProtocolID
			, p_pszDataSetName, p_pszWellIndexes)

	def GetDataSetResults(self, p_pszProtocolID=defaultNamedNotOptArg, p_pszDataSetName=defaultNamedNotOptArg, p_pszResultIndexes=defaultNamedNotOptArg):
		'method GetDataSetResults'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(50, LCID, 1, (8, 0), ((8, 0), (8, 0), (8, 0)),p_pszProtocolID
			, p_pszDataSetName, p_pszResultIndexes)

	def GetFileExportNames(self, AutoExecuteOnly=defaultNamedNotOptArg, pvNames=defaultNamedNotOptArg):
		'method GetFileExportNames'
		return self._oleobj_.InvokeTypes(29, LCID, 1, (24, 0), ((11, 0), (16396, 0)),AutoExecuteOnly
			, pvNames)

	def GetImageFolderPaths(self, pvPaths=defaultNamedNotOptArg):
		'method GetImageFolderPaths'
		return self._oleobj_.InvokeTypes(40, LCID, 1, (24, 0), ((16396, 0),),pvPaths
			)

	def GetModifiableProcedure(self):
		'method GetModifiableProcedure'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(56, LCID, 1, (8, 0), (),)

	def GetMovieExportNames(self, AutoExecuteOnly=defaultNamedNotOptArg, pvNames=defaultNamedNotOptArg):
		'method GetMovieExportNames'
		return self._oleobj_.InvokeTypes(42, LCID, 1, (24, 0), ((11, 0), (16396, 0)),AutoExecuteOnly
			, pvNames)

	def GetOLEMetaData(self, pvMetaData=defaultNamedNotOptArg):
		'method GetOLEMetaData'
		return self._oleobj_.InvokeTypes(67, LCID, 1, (24, 0), ((16396, 0),),pvMetaData
			)

	def GetPictureExportNames(self, AutoExecuteOnly=defaultNamedNotOptArg, pvNames=defaultNamedNotOptArg):
		'method GetPictureExportNames'
		return self._oleobj_.InvokeTypes(41, LCID, 1, (24, 0), ((11, 0), (16396, 0)),AutoExecuteOnly
			, pvNames)

	def GetPowerExportNames(self, AutoExecuteOnly=defaultNamedNotOptArg, pvNames=defaultNamedNotOptArg):
		'method GetPowerExportNames'
		return self._oleobj_.InvokeTypes(30, LCID, 1, (24, 0), ((11, 0), (16396, 0)),AutoExecuteOnly
			, pvNames)

	def GetProcedure(self):
		'method GetProcedure'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(34, LCID, 1, (8, 0), (),)

	def GetRawData(self, pDataSetName=defaultNamedNotOptArg, pvRow=defaultNamedNotOptArg, pvColumn=defaultNamedNotOptArg, pvKineticIndex=defaultNamedNotOptArg
			, pvWavelengthIndex=defaultNamedNotOptArg, pvHorizontalIndex=defaultNamedNotOptArg, pvVerticalIndex=defaultNamedNotOptArg, pvValue=defaultNamedNotOptArg, pvPrimaryStatus=defaultNamedNotOptArg
			, pvSecondaryStatus=defaultNamedNotOptArg):
		'method GetRawData'
		return self._oleobj_.InvokeTypes(25, LCID, 1, (3, 0), ((16392, 0), (16396, 0), (16396, 0), (16396, 0), (16396, 0), (16396, 0), (16396, 0), (16396, 0), (16396, 0), (16396, 0)),pDataSetName
			, pvRow, pvColumn, pvKineticIndex, pvWavelengthIndex, pvHorizontalIndex
			, pvVerticalIndex, pvValue, pvPrimaryStatus, pvSecondaryStatus)

	def GetReportNames(self, AutoExecuteOnly=defaultNamedNotOptArg, pvNames=defaultNamedNotOptArg):
		'method GetReportNames'
		return self._oleobj_.InvokeTypes(28, LCID, 1, (24, 0), ((11, 0), (16396, 0)),AutoExecuteOnly
			, pvNames)

	def GetSampleDescription(self):
		'method GetSampleDescription'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(26, LCID, 1, (8, 0), (),)

	def ImportSampleIDs(self, ImportFilePath=defaultNamedNotOptArg):
		'method ImportSampleIDs'
		return self._oleobj_.InvokeTypes(16, LCID, 1, (24, 0), ((8, 0),),ImportFilePath
			)

	def KeepPlateInAfterRead(self):
		'method KeepPlateInAfterRead'
		return self._oleobj_.InvokeTypes(46, LCID, 1, (24, 0), (),)

	def MovieExport(self, p_pszBuilderName=defaultNamedNotOptArg):
		'method MovieExport'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(44, LCID, 1, (8, 0), ((8, 0),),p_pszBuilderName
			)

	def PictureExport(self, p_pszBuilderName=defaultNamedNotOptArg):
		'method PictureExport'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(43, LCID, 1, (8, 0), ((8, 0),),p_pszBuilderName
			)

	def PowerExport(self, ExportFilePath=defaultNamedNotOptArg):
		'method PowerExport'
		return self._oleobj_.InvokeTypes(12, LCID, 1, (24, 0), ((8, 0),),ExportFilePath
			)

	def PowerExportEx(self, p_pszBuilderName=defaultNamedNotOptArg, p_pszfilePath=defaultNamedNotOptArg):
		'method PowerExportEx'
		return self._oleobj_.InvokeTypes(33, LCID, 1, (24, 0), ((8, 0), (8, 0)),p_pszBuilderName
			, p_pszfilePath)

	def PrintReport(self, ShowPrintDlg=defaultNamedNotOptArg):
		'method PrintReport'
		return self._oleobj_.InvokeTypes(15, LCID, 1, (24, 0), ((11, 0),),ShowPrintDlg
			)

	def PrintReportEx(self, p_pszBuilderName=defaultNamedNotOptArg, bShowPrintDlg=defaultNamedNotOptArg):
		'method PrintReportEx'
		return self._oleobj_.InvokeTypes(31, LCID, 1, (24, 0), ((8, 0), (3, 0)),p_pszBuilderName
			, bShowPrintDlg)

	def QCExport(self):
		'method QCExport'
		return self._oleobj_.InvokeTypes(21, LCID, 1, (24, 0), (),)

	def ResumeRead(self):
		'method ResumeRead'
		ret = self._oleobj_.InvokeTypes(9, LCID, 1, (9, 0), (),)
		if ret is not None:
			ret = Dispatch(ret, 'ResumeRead', None)
		return ret

	def ResumeReadEx(self, p_bRotate180Deg=defaultNamedNotOptArg):
		'method ResumeReadEx'
		ret = self._oleobj_.InvokeTypes(37, LCID, 1, (9, 0), ((11, 0),),p_bRotate180Deg
			)
		if ret is not None:
			ret = Dispatch(ret, 'ResumeReadEx', None)
		return ret

	def ResumeSimulatedRead(self):
		'method ResumeSimulatedRead'
		ret = self._oleobj_.InvokeTypes(10, LCID, 1, (9, 0), (),)
		if ret is not None:
			ret = Dispatch(ret, 'ResumeSimulatedRead', None)
		return ret

	def SetChannelBCValues(self, ChannelName=defaultNamedNotOptArg, BrightnessValue=defaultNamedNotOptArg, ContrastValue=defaultNamedNotOptArg):
		'method SetChannelBCValues'
		return self._oleobj_.InvokeTypes(62, LCID, 1, (24, 0), ((8, 0), (3, 0), (3, 0)),ChannelName
			, BrightnessValue, ContrastValue)

	def SetDataMaskedState(self, MaskRequestsXML=defaultNamedNotOptArg):
		'method SetDataMaskedState'
		return self._oleobj_.InvokeTypes(60, LCID, 1, (24, 0), ((8, 0),),MaskRequestsXML
			)

	def SetModifiableDataReduction(self, ProtocolID=defaultNamedNotOptArg, ModifiableDataReductionXML=defaultNamedNotOptArg):
		'method SetModifiableDataReduction'
		return self._oleobj_.InvokeTypes(59, LCID, 1, (24, 0), ((8, 0), (8, 0)),ProtocolID
			, ModifiableDataReductionXML)

	def SetModifiableProcedure(self, ModifiableProcedureXML=defaultNamedNotOptArg):
		'method SetModifiableProcedure'
		return self._oleobj_.InvokeTypes(57, LCID, 1, (24, 0), ((8, 0),),ModifiableProcedureXML
			)

	def SetOLEMetaData(self, pvMetaData=defaultNamedNotOptArg):
		'method SetOLEMetaData'
		return self._oleobj_.InvokeTypes(68, LCID, 1, (24, 0), ((16396, 0),),pvMetaData
			)

	def SetPartialPlate(self, p_pszPartialPlateXML=defaultNamedNotOptArg):
		'method SetPartialPlate'
		return self._oleobj_.InvokeTypes(45, LCID, 1, (24, 0), ((8, 0),),p_pszPartialPlateXML
			)

	def SetProcedure(self, ProcedureXML=defaultNamedNotOptArg):
		'method SetProcedure'
		return self._oleobj_.InvokeTypes(39, LCID, 1, (24, 0), ((8, 0),),ProcedureXML
			)

	def SetSampleDescription(self, SampleDescriptionXML=defaultNamedNotOptArg):
		'method SetSampleDescription'
		return self._oleobj_.InvokeTypes(27, LCID, 1, (24, 0), ((8, 0),),SampleDescriptionXML
			)

	def StartRead(self):
		'method StartRead'
		ret = self._oleobj_.InvokeTypes(6, LCID, 1, (9, 0), (),)
		if ret is not None:
			ret = Dispatch(ret, 'StartRead', None)
		return ret

	def StartReadEx(self, p_bRotate180Deg=defaultNamedNotOptArg):
		'method StartReadEx'
		ret = self._oleobj_.InvokeTypes(35, LCID, 1, (9, 0), ((11, 0),),p_bRotate180Deg
			)
		if ret is not None:
			ret = Dispatch(ret, 'StartReadEx', None)
		return ret

	def StartReadFromFile(self, FilePath=defaultNamedNotOptArg, ListSeparator=defaultNamedNotOptArg):
		'method StartReadFromFile'
		ret = self._oleobj_.InvokeTypes(8, LCID, 1, (9, 0), ((8, 0), (8, 0)),FilePath
			, ListSeparator)
		if ret is not None:
			ret = Dispatch(ret, 'StartReadFromFile', None)
		return ret

	def StartSimulatedRead(self):
		'method StartSimulatedRead'
		ret = self._oleobj_.InvokeTypes(7, LCID, 1, (9, 0), (),)
		if ret is not None:
			ret = Dispatch(ret, 'StartSimulatedRead', None)
		return ret

	def TryCellularAnalysisStep(self, CellularAnalysisStepXML=defaultNamedNotOptArg, ImageReferenceXML=defaultNamedNotOptArg, PictureExportBuilderName=defaultNamedNotOptArg, OutputPicturePathname=defaultNamedNotOptArg
			, p_ObjectCount=defaultNamedNotOptArg):
		'method TryCellularAnalysisStep'
		return self._oleobj_.InvokeTypes(63, LCID, 1, (24, 0), ((8, 0), (8, 0), (8, 0), (8, 0), (16396, 0)),CellularAnalysisStepXML
			, ImageReferenceXML, PictureExportBuilderName, OutputPicturePathname, p_ObjectCount)

	def TryCellularAnalysisStepEx(self, CellularAnalysisStepXML=defaultNamedNotOptArg, ImageReferenceXML=defaultNamedNotOptArg, PictureExportBuilderName=defaultNamedNotOptArg, OutputPicturePathname=defaultNamedNotOptArg
			, p_ObjectCounts=defaultNamedNotOptArg):
		'method TryCellularAnalysisStepEx'
		return self._oleobj_.InvokeTypes(71, LCID, 1, (24, 0), ((8, 0), (8, 0), (8, 0), (8, 0), (16396, 0)),CellularAnalysisStepXML
			, ImageReferenceXML, PictureExportBuilderName, OutputPicturePathname, p_ObjectCounts)

	def ValidateProcedure(self, p_bRotate180Deg=defaultNamedNotOptArg, pbstrRejectionCause=defaultNamedNotOptArg):
		'method ValidateProcedure'
		return self._oleobj_.InvokeTypes(36, LCID, 1, (3, 0), ((11, 0), (16392, 0)),p_bRotate180Deg
			, pbstrRejectionCause)

	_prop_map_get_ = {
		"Barcode": (2, 2, (8, 0), (), "Barcode", None),
		"CalculationStatus": (47, 2, (3, 0), (), "CalculationStatus", None),
		"Comment": (38, 2, (8, 0), (), "Comment", None),
		"DisKinInterval": (18, 2, (3, 0), (), "DisKinInterval", None),
		"DisKinRuntime": (19, 2, (3, 0), (), "DisKinRuntime", None),
		"DiscontinuousKinetic": (17, 2, (11, 0), (), "DiscontinuousKinetic", None),
		"ID": (1, 2, (8, 0), (), "ID", None),
		"Labware": (53, 2, (11, 0), (), "Labware", None),
		"MaxColumns": (23, 2, (3, 0), (), "MaxColumns", None),
		"MaxRows": (22, 2, (3, 0), (), "MaxRows", None),
		"MaxVessels": (54, 2, (3, 0), (), "MaxVessels", None),
		"Name": (3, 2, (8, 0), (), "Name", None),
		"PlateType": (4, 2, (3, 0), (), "PlateType", None),
		"ReadStatus": (5, 2, (3, 0), (), "ReadStatus", None),
		"SampleCount": (20, 2, (3, 0), (), "SampleCount", None),
		"ValidationUIEnabled": (70, 2, (11, 0), (), "ValidationUIEnabled", None),
	}
	_prop_map_put_ = {
		"Barcode" : ((2, LCID, 4, 0),()),
		"CalculationStatus" : ((47, LCID, 4, 0),()),
		"Comment" : ((38, LCID, 4, 0),()),
		"DisKinInterval" : ((18, LCID, 4, 0),()),
		"DisKinRuntime" : ((19, LCID, 4, 0),()),
		"DiscontinuousKinetic" : ((17, LCID, 4, 0),()),
		"ID" : ((1, LCID, 4, 0),()),
		"Labware" : ((53, LCID, 4, 0),()),
		"MaxColumns" : ((23, LCID, 4, 0),()),
		"MaxRows" : ((22, LCID, 4, 0),()),
		"MaxVessels" : ((54, LCID, 4, 0),()),
		"Name" : ((3, LCID, 4, 0),()),
		"PlateType" : ((4, LCID, 4, 0),()),
		"ReadStatus" : ((5, LCID, 4, 0),()),
		"SampleCount" : ((20, LCID, 4, 0),()),
		"ValidationUIEnabled" : ((70, LCID, 4, 0),()),
	}
	def __iter__(self):
		"Return a Python iterator for this object"
		try:
			ob = self._oleobj_.InvokeTypes(-4,LCID,3,(13, 10),())
		except pythoncom.error:
			raise TypeError("This object does not support enumeration")
		return win32com.client.util.Iterator(ob, None)

class IPlateExportMonitor(DispatchBaseClass):
	CLSID = IID('{28B75D71-F78A-417A-A9E5-2D7F9B16D3F6}')
	coclass_clsid = IID('{B5CC44DB-5EAA-49A9-A6E0-72F23C067141}')

	def GetProgressDetail(self, p_psCompleted=defaultNamedNotOptArg, p_psTotal=defaultNamedNotOptArg):
		'method GetProgressDetail'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(3, LCID, 1, (8, 0), ((16396, 0), (16396, 0)),p_psCompleted
			, p_psTotal)

	def RequestAbortion(self):
		'method RequestAbortion'
		return self._oleobj_.InvokeTypes(2, LCID, 1, (24, 0), (),)

	_prop_map_get_ = {
		"ExportInProgress": (1, 2, (11, 0), (), "ExportInProgress", None),
	}
	_prop_map_put_ = {
		"ExportInProgress" : ((1, LCID, 4, 0),()),
	}
	def __iter__(self):
		"Return a Python iterator for this object"
		try:
			ob = self._oleobj_.InvokeTypes(-4,LCID,3,(13, 10),())
		except pythoncom.error:
			raise TypeError("This object does not support enumeration")
		return win32com.client.util.Iterator(ob, None)

class IPlateReadMonitor(DispatchBaseClass):
	CLSID = IID('{8C5C803A-D465-4DF3-B996-663EC0D64A5C}')
	coclass_clsid = IID('{DC9E09F7-1EBD-4BD9-B2A3-7F41FA3CC782}')

	def GetErrorCode(self, ErrorIndex=defaultNamedNotOptArg):
		'method GetErrorCode'
		return self._oleobj_.InvokeTypes(3, LCID, 1, (3, 0), ((3, 0),),ErrorIndex
			)

	def GetErrorMessage(self, ErrorIndex=defaultNamedNotOptArg):
		'method GetErrorMessage'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(4, LCID, 1, (8, 0), ((3, 0),),ErrorIndex
			)

	def GetReaderError(self, ErrorIndex=defaultNamedNotOptArg):
		'method GetReaderError'
		return self._oleobj_.InvokeTypes(5, LCID, 1, (3, 0), ((3, 0),),ErrorIndex
			)

	_prop_map_get_ = {
		"BulbWarmupRemainingSeconds": (6, 2, (3, 0), (), "BulbWarmupRemainingSeconds", None),
		"ErrorsCount": (2, 2, (3, 0), (), "ErrorsCount", None),
		"ReadInProgress": (1, 2, (11, 0), (), "ReadInProgress", None),
	}
	_prop_map_put_ = {
		"BulbWarmupRemainingSeconds" : ((6, LCID, 4, 0),()),
		"ErrorsCount" : ((2, LCID, 4, 0),()),
		"ReadInProgress" : ((1, LCID, 4, 0),()),
	}
	def __iter__(self):
		"Return a Python iterator for this object"
		try:
			ob = self._oleobj_.InvokeTypes(-4,LCID,3,(13, 10),())
		except pythoncom.error:
			raise TypeError("This object does not support enumeration")
		return win32com.client.util.Iterator(ob, None)

class IPlates(DispatchBaseClass):
	CLSID = IID('{CA9E1D87-7011-4322-94A7-E95282725B6A}')
	coclass_clsid = IID('{CBBDA8E5-A721-40D9-B677-D12CA91721CC}')

	def Add(self):
		'method Add'
		ret = self._oleobj_.InvokeTypes(4, LCID, 1, (9, 0), (),)
		if ret is not None:
			ret = Dispatch(ret, 'Add', None)
		return ret

	def GetPlate(self, Index=defaultNamedNotOptArg):
		'method GetPlate'
		ret = self._oleobj_.InvokeTypes(3, LCID, 1, (9, 0), ((3, 0),),Index
			)
		if ret is not None:
			ret = Dispatch(ret, 'GetPlate', None)
		return ret

	_prop_map_get_ = {
		"CalibrationCount": (2, 2, (3, 0), (), "CalibrationCount", None),
		"Count": (1, 2, (3, 0), (), "Count", None),
	}
	_prop_map_put_ = {
		"CalibrationCount" : ((2, LCID, 4, 0),()),
		"Count" : ((1, LCID, 4, 0),()),
	}
	def __iter__(self):
		"Return a Python iterator for this object"
		try:
			ob = self._oleobj_.InvokeTypes(-4,LCID,3,(13, 10),())
		except pythoncom.error:
			raise TypeError("This object does not support enumeration")
		return win32com.client.util.Iterator(ob, None)
	#This class has Count() property - allow len(ob) to provide this
	def __len__(self):
		return self._ApplyTypes_(*(1, 2, (3, 0), (), "Count", None))
	#This class has a __len__ - this is needed so 'if object:' always returns TRUE.
	def __nonzero__(self):
		return True

class ISystemTestMonitor(DispatchBaseClass):
	CLSID = IID('{E4308F75-5180-43C8-BB1E-29043281E2AF}')
	coclass_clsid = IID('{3077BF18-B3A4-4C7D-8D97-63AA7DA5333E}')

	def GetSystemTestResults(self):
		'method GetSystemTestResults'
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(3, LCID, 1, (8, 0), (),)

	_prop_map_get_ = {
		"TestInProgress": (1, 2, (11, 0), (), "TestInProgress", None),
		"TestPassed": (2, 2, (11, 0), (), "TestPassed", None),
	}
	_prop_map_put_ = {
		"TestInProgress" : ((1, LCID, 4, 0),()),
		"TestPassed" : ((2, LCID, 4, 0),()),
	}
	def __iter__(self):
		"Return a Python iterator for this object"
		try:
			ob = self._oleobj_.InvokeTypes(-4,LCID,3,(13, 10),())
		except pythoncom.error:
			raise TypeError("This object does not support enumeration")
		return win32com.client.util.Iterator(ob, None)

from win32com.client import CoClassBaseClass
# This CoClass is known by the name 'Gen5.Application'
class Application(CoClassBaseClass): # A CoClass
	CLSID = IID('{9B3B05F7-143C-4DAF-93D8-726A7BA3321A}')
	coclass_sources = [
	]
	coclass_interfaces = [
		IApplication,
	]
	default_interface = IApplication

class Experiment(CoClassBaseClass): # A CoClass
	CLSID = IID('{F4024CA7-7362-4FC8-B33F-F1F2BD58788F}')
	coclass_sources = [
	]
	coclass_interfaces = [
		IExperiment,
	]
	default_interface = IExperiment

class MonoCalTestMonitor(CoClassBaseClass): # A CoClass
	CLSID = IID('{9EFA02E9-6776-44B3-B193-62A360356658}')
	coclass_sources = [
	]
	coclass_interfaces = [
		IMonoCalTestMonitor,
	]
	default_interface = IMonoCalTestMonitor

class Plate(CoClassBaseClass): # A CoClass
	CLSID = IID('{DA2A9CC3-FE47-441B-BD0C-B482015A525F}')
	coclass_sources = [
	]
	coclass_interfaces = [
		IPlate,
	]
	default_interface = IPlate

class PlateExportMonitor(CoClassBaseClass): # A CoClass
	CLSID = IID('{B5CC44DB-5EAA-49A9-A6E0-72F23C067141}')
	coclass_sources = [
	]
	coclass_interfaces = [
		IPlateExportMonitor,
	]
	default_interface = IPlateExportMonitor

class PlateReadMonitor(CoClassBaseClass): # A CoClass
	CLSID = IID('{DC9E09F7-1EBD-4BD9-B2A3-7F41FA3CC782}')
	coclass_sources = [
	]
	coclass_interfaces = [
		IPlateReadMonitor,
	]
	default_interface = IPlateReadMonitor

class Plates(CoClassBaseClass): # A CoClass
	CLSID = IID('{CBBDA8E5-A721-40D9-B677-D12CA91721CC}')
	coclass_sources = [
	]
	coclass_interfaces = [
		IPlates,
	]
	default_interface = IPlates

class SystemTestMonitor(CoClassBaseClass): # A CoClass
	CLSID = IID('{3077BF18-B3A4-4C7D-8D97-63AA7DA5333E}')
	coclass_sources = [
	]
	coclass_interfaces = [
		ISystemTestMonitor,
	]
	default_interface = ISystemTestMonitor

RecordMap = {
}

CLSIDToClassMap = {
	'{56BBF0D8-DDE6-4F25-BEE1-F3A7CC516531}' : IApplication,
	'{9B3B05F7-143C-4DAF-93D8-726A7BA3321A}' : Application,
	'{21258B83-0AAE-43F7-AFBC-89C680D51C6F}' : IExperiment,
	'{F4024CA7-7362-4FC8-B33F-F1F2BD58788F}' : Experiment,
	'{CA9E1D87-7011-4322-94A7-E95282725B6A}' : IPlates,
	'{CBBDA8E5-A721-40D9-B677-D12CA91721CC}' : Plates,
	'{450BF0EA-F196-4D18-AD0D-DB39D2E6E08B}' : IPlate,
	'{DA2A9CC3-FE47-441B-BD0C-B482015A525F}' : Plate,
	'{8C5C803A-D465-4DF3-B996-663EC0D64A5C}' : IPlateReadMonitor,
	'{DC9E09F7-1EBD-4BD9-B2A3-7F41FA3CC782}' : PlateReadMonitor,
	'{0DABEE12-4422-4FE2-8D4B-F510CA25ECD7}' : IMonoCalTestMonitor,
	'{9EFA02E9-6776-44B3-B193-62A360356658}' : MonoCalTestMonitor,
	'{E4308F75-5180-43C8-BB1E-29043281E2AF}' : ISystemTestMonitor,
	'{3077BF18-B3A4-4C7D-8D97-63AA7DA5333E}' : SystemTestMonitor,
	'{28B75D71-F78A-417A-A9E5-2D7F9B16D3F6}' : IPlateExportMonitor,
	'{B5CC44DB-5EAA-49A9-A6E0-72F23C067141}' : PlateExportMonitor,
}
CLSIDToPackageMap = {}
win32com.client.CLSIDToClass.RegisterCLSIDsFromDict( CLSIDToClassMap )
VTablesToPackageMap = {}
VTablesToClassMap = {
}


NamesToIIDMap = {
	'IApplication' : '{56BBF0D8-DDE6-4F25-BEE1-F3A7CC516531}',
	'IExperiment' : '{21258B83-0AAE-43F7-AFBC-89C680D51C6F}',
	'IPlates' : '{CA9E1D87-7011-4322-94A7-E95282725B6A}',
	'IPlate' : '{450BF0EA-F196-4D18-AD0D-DB39D2E6E08B}',
	'IPlateReadMonitor' : '{8C5C803A-D465-4DF3-B996-663EC0D64A5C}',
	'IMonoCalTestMonitor' : '{0DABEE12-4422-4FE2-8D4B-F510CA25ECD7}',
	'ISystemTestMonitor' : '{E4308F75-5180-43C8-BB1E-29043281E2AF}',
	'IPlateExportMonitor' : '{28B75D71-F78A-417A-A9E5-2D7F9B16D3F6}',
}

win32com.client.constants.__dicts__.append(constants.__dict__)




import logging

from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.cytation_pb2 import Command, Config

from .driver import CytationDriver
import argparse 

class CytationServer(ToolServer):
    toolType = "cytation"

    def __init__(self) -> None:
        super().__init__()
        self.driver: CytationDriver

    def _configure(self, config: Config) -> None:
        self.config = config
        if self.driver:
            self.driver.close()
        self.driver = CytationDriver(
            protocol_dir=config.protocol_dir,
            experiment_dir=config.experiment_dir,
            reader_type=config.reader_type,
        )
        self.driver.verify_reader_communication()

    def OpenCarrier(self, params: Command.OpenCarrier) -> None:
        self.driver.open_carrier()

    def CloseCarrier(self, params: Command.CloseCarrier) -> None:
        self.driver.close_carrier()

    def StartRead(self, params: Command.StartRead) -> None:
        self.driver.start_read(
            protocol_file=params.protocol_file,
            experiment_name=params.experiment_name,
            well_addresses=[addy for addy in params.well_addresses],
        )

    def EstimateOpenCarrier(self, params: Command.OpenCarrier) -> int:
        return 2

    def EstimateCloseCarrier(self, params: Command.CloseCarrier) -> int:
        return 2

    def EstimateStartRead(self, params: Command.StartRead) -> int:
        return 5


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
         raise RuntimeWarning("Port must be provided...")
    serve(CytationServer(), str(args.port))



import typing as t
import logging
from tools.base_server import ABCToolDriver
import requests
import time
from typing import Optional 
from PIL import Image
import io
import os 

class Ot2Driver(ABCToolDriver):
    def __init__(self, robot_ip: str, robot_port: int = 31950) -> None:
        self.robot_ip: str = robot_ip
        self.robot_port: int = robot_port

        self.run_id: Optional[str] = None

        self.base_url = f"http://{robot_ip}:{robot_port}"
        self.headers = {"Opentrons-Version": "2"}

    def ping(self) -> None:
        response = requests.get(
            url=f"{self.base_url}/health",
            headers=self.headers,
        )

        if response.status_code != 200:
            raise Exception("ping failed")
        else:
            logging.info("ping finished")
    
    def toggle_light(self) -> None:
        # Toggle opentrons light
        light_status_response = requests.get(
            url=f"{self.base_url}/robot/lights",
            headers=self.headers,
        )
        if not light_status_response.ok:
            raise Exception(f"get light status failed: {light_status_response.text}")
        light_status: bool = light_status_response.json()['on']

        toggle_light_response = requests.post(
            url=f"{self.base_url}/robot/lights",
            headers=self.headers,
            json={"on": not light_status}
        )
        if not toggle_light_response.ok:
            raise Exception("toggle light failed")


    def schedule_run(self, protocol_file: str) -> str:
        with open(protocol_file, 'rb') as F:
            upload_protocol_response = requests.post(
                url=f"{self.base_url}/protocols",
                files={
                    "files": ('json_program.py', F, "text/x-python-script"),
                },
                headers=self.headers,
            )
        try:
            protocol_id = upload_protocol_response.json()['data']['id']
        except KeyError:
            raise Exception(f"protocol upload failed: {upload_protocol_response.text}")

        # Create a run for the protocol
        create_run_response = requests.post(
            url=f"{self.base_url}/runs",
            json={
                "data": {
                    "protocolId": protocol_id
                }
            },
            headers=self.headers,
        )
        run_id: str = create_run_response.json()['data']['id']
        self.run_id = run_id
        logging.info(f'Created run with id: {self.run_id}')

        # Start run
        start_run_response = requests.post(
            url=f"{self.base_url}/runs/{self.run_id}/actions",
            headers=self.headers,
            json={"data": {"actionType": "play"}}
        )
        if not start_run_response.ok:
            raise Exception("run failed to start")
        
        return run_id

    def pause_protocol(self) -> None:
        if self.run_id:
            pause_run_response = requests.post(
                url=f"{self.base_url}/runs/{self.run_id}/actions",
                headers=self.headers,
                json={"data": {"actionType": "pause"}}
            )
            if not pause_run_response.ok:
                raise Exception("run failed to pause")
    
    def resume_protocol(self) -> None:
        if self.run_id:
            resume_run_response = requests.post(
                url=f"{self.base_url}/runs/{self.run_id}/actions",
                headers=self.headers,
                json={"data": {"actionType": "play"}}
            )
            if not resume_run_response.ok:
                raise Exception("run failed to resume")
    
    def cancel_protocol(self) -> None:
        if self.run_id:
            stop_run_response = requests.post(
                url=f"{self.base_url}/runs/{self.run_id}/actions",
                headers=self.headers,
                json={"data": {"actionType": "stop"}}
            )
            if not stop_run_response.ok:
                raise Exception("run failed to stop")
            delete_run_response = requests.delete(
                url=f"{self.base_url}/runs/{self.run_id}",
                headers=self.headers,
            )
            if not delete_run_response.ok:
                raise Exception("run failed to delete")
            self.run_id = None
    
    def get_run(self, run_id: str) -> t.Any:
        get_run_response = requests.get(
            url=f"{self.base_url}/runs/{run_id}",
            headers=self.headers,
            timeout=30
        )
        if not get_run_response.ok:
            raise Exception(f"get run failed: {get_run_response.text}")
        return get_run_response.json()

    def wait_for_command(self, run_id: str, timeout: int = 1800) -> None:
        seconds: int = 0
        run_status: str = self.get_run(run_id)['data']['status']
        while run_status != 'succeeded':
            if seconds > timeout:
                raise TimeoutError(
                    f"OT Run {run_id} did not complete in {timeout} seconds"
                )
            run_status = self.get_run(run_id)['data']['status']
            logging.info(f"OT Run {run_id} status: {run_status}")
            if run_status in ['stopped', 'failed', 'blocked-by-open-door']:
                raise Exception(f"OT Run {run_id} failed")
            if run_status in ['succeeded']:
                return None
            time.sleep(1)
            seconds += 1

    def start_protocol(self, protocol_file: str) -> None:
        logging.info(f"starting protocol {protocol_file}")
        run_id = self.schedule_run(protocol_file=protocol_file)
        self.wait_for_command(run_id)
        time.sleep(2)
    
    def take_picture(self, name:str, directory:str) -> None:
        response = requests.post(
            url=f"{self.base_url}/camera/picture",
            headers=self.headers,
        )
        if not response.ok:
            raise Exception("Failed to take picture")
        
        image = Image.open(io.BytesIO(response.content))
        rotated_image = image.rotate(180)
        if not os.path.exists(directory):
            os.makedirs(directory)
        file_path : str = os.path.join(directory,name)
        rotated_image.save(file_path, format="JPEG")




import typing as t
import logging
import time
import traceback
from tools.base_server import ABCToolDriver
from typing import Union, Optional
import serial

MAX_RETRIES = 5
# How long to wait before firing another scanned barcode to the listeners.
SCANNED_BARCODE_THROTTLE = 0.25
READ_TIMEOUT = 1


def try_utf_decode(data: Union[str,bytes]) -> str:
    if isinstance(data, str):
        return data
    data_string = ""
    try:
        data_string = data.decode("utf-8")
    except Exception:
        raise serial.SerialException(f"error decoding to utf8 string: {data!r}")
    return data_string


def serial_read(serial_port: serial.Serial) -> str:
    serial_port.reset_input_buffer()
    reply = serial_port.read_until(expected=b"\n")
    reply_string = try_utf_decode(reply)
    if reply_string == "":
        logging.warning("Dataman 70 returned empty response")
        return ""
    # The dataman 70 returns malformed responses every once in a while. Just ignore them.
    if not (reply_string[-1] == "\n" and reply_string[-2] == "\r"):
        logging.warning(f"Dataman 70 returned malformed response {reply_string}")
        return ""
    return reply_string[0:-2]


class Dataman70Driver(ABCToolDriver):
    def __init__(self, com_port: str) -> None:
        self.debug_output: bool = True
        self.serial_timeout: int = 2
        # The last barcode received.
        self.last_barcode: str = ""
        # The time the last barcode was received.
        self.last_barcode_time: Optional[float] = None
        # Listeners that will fire whenever a new barcode is received.
        self.barcode_listeners = []

        # Whether the driver is live.
        # If the driver cannot receive a signal, it will set live to False.
        # If the driver is not live, the reader_thread will halt.
        # If the driver is not live, it will be marked as disconnected via the healthcheck.
        self.live: bool = True

        # NOTE: We've seen issues where when the serial is first connected,
        # creating the serial port can take 10 seconds or more.
        # This only occurs the first time.
        self.serial_port: serial.Serial = serial.Serial(
            com_port, write_timeout=1, timeout=READ_TIMEOUT
        )

        # self.reader_thread: threading.Thread = threading.Thread(target=self.read_dataman_output)
        # self.reader_thread.daemon = True
        # self.reader_thread.start()

        # Responses from the dataman 70.
        self.responses: dict[str, t.Any] = {}

    def close(self) -> None:
        self.live = False
        self.serial_port.close()

    def process_dataman_output(self, output: str) -> None:
        if output == "":
            return

        if len(output) == 1:
            self.responses["current_trigger_type"] = output
            return

        if (
            self.last_barcode_time is None
            or time.time() - self.last_barcode_time > SCANNED_BARCODE_THROTTLE
        ):
            if self.debug_output:
                logging.info(f"Dataman detected barcode... {output}")
            self.last_barcode = output
            self.last_barcode_time = time.time()

            try:
                for listener in self.barcode_listeners:
                    listener(self.last_barcode)
            except Exception:
                logging.error(traceback.format_exc())

    def read_dataman_output(self) -> None:
        self.last_barcode = ""
        self.last_barcode_time = time.time()
        retries = 0
        while self.live:
            try:
                # Read from the Dataman.
                output = serial_read(self.serial_port)
                # # Reset retries if read is successful.
                # retries = 0

                self.process_dataman_output(output)

            except serial.SerialException:
                if self.live:
                    logging.error(traceback.format_exc())
                    retries += 1
                    time.sleep(1)

            except TypeError:
                # If not live anymore, we expect a error from read() in serial/serialposix.py.
                if not self.live:
                    pass
                else:
                    raise
            retries += 1
            # If we receive multiple errors attempting to receive data, set to disconnected.
            # We assume that if the Dataman is disconnected, this will trigger.
            if retries > MAX_RETRIES:
                self.live = False
                # read_dataman_output happens in its own thread, NOT inside the control_command decorator.
                # Oo we cannot simply raise Exception to get the instrument marked as disconnected.
                # To remedy this, we check self.live in the healthcheck.
                logging.error(
                    "Could not get valid data after multiple retries. Marking instrument as disconnected"
                )

    def power_on(self) -> None:
        self.serial_port.write(b"||>SET TRIGGER.TYPE 1\r\n")
        logging.info("Set trigger type to 1.")

    def power_off(self) -> None:
        self.serial_port.write(b"||>SET TRIGGER.TYPE 0\r\n")
        logging.info("Set trigger type to 0.")

    """
        Wait for a response, which is filled by process_dataman_output.
    """

    def _wait_for_response(self, response_type: str) -> None:
        response_retry_interval = 0.2
        max_retries = self.serial_timeout / response_retry_interval
        retries = 0
        while retries < max_retries and self.responses[response_type] is None:
            time.sleep(response_retry_interval)
            retries += 1

        # If no response is recorded, raise an error.
        if self.responses[response_type] is None:
            raise Exception(f"Timed out waiting for response {response_type}")

    # Get the on state by querying the barcode reader.
    def is_on(self) -> bool:
        # Send command
        self.responses["current_trigger_type"] = None
        self.serial_port.write(b"||>GET TRIGGER.TYPE\r\n")

        self._wait_for_response("current_trigger_type")
        if self.responses["current_trigger_type"] == "1":
            return True
        return False

    def read_barcode(self) -> str:
        if not self.live:
            raise Exception("Dataman 70 is not live")

        for retry in range(MAX_RETRIES):
            barcode = serial_read(self.serial_port)
            if barcode != "":
                return barcode

        return ""

    def register_barcode_listener(self, fn: t.Callable[[str], t.Any]) -> None:
        self.barcode_listeners.append(fn)



from tools.comms.tcpip import TcpIp
from tools.base_server import ABCToolDriver
import time 
import logging 
from typing import Optional 

COMMANDS = {
    "home":"h",
    "load":"l",
    "unload":"u",
    "retract":"rt",
    "abort":"a",
    "scan":"bs",
    "set_plate_height":"sph",
    "set_plate_dimensions":"spd",
    "set_plate_thickness":"spt",
    # "status":"s", #Do not use this command for rapid polling bc it communicates with the aplifiers within the machine.
    "varstatus":"vs", #this command is appropiate for use withing driver. 
    "spin":"sp",
    "manual":"m", #set to manual mode. ,
    "get_dimensions":"dimstatus",
    "clear_error":"cba"
}
class MicroServeDriver(ABCToolDriver):
    toolType = "microserve"
    
    def __init__(self, ip:str, port: int) -> None:
        self.tcp : Optional[TcpIp] = None
        self.blocking :bool = True
        self.port : int = port
        self.ip : str = ip
        self.error : Optional[str] = None
    
    def connect(self) -> None:
        logging.info(f"Connecting to microserve at ip: {self.ip} and port {self.port}")
        try:
            self.tcp = TcpIp(self.ip, self.port)
            self.tcp.connect()
        except Exception as e:
            logging.info(f"Failed to connect to MicroServe: {e}")
            self.error = str(e)
            raise ValueError(f"Failed to connect to MicroServe: {e}")

    def get_status(self) -> str:
        return  self.send_command(COMMANDS["varstatus"])
    
    def is_homed(self) -> bool:
        logging.info("Checking if microserve is homed")
        response = self.get_status()
        logging.info(f"Home status {response}")
        time.sleep(0.1)
        if "OK! status homed" in response:
            logging.info("Microserve is already homed")
            return True
        else:
            logging.info("Microserve is not homed")
            return False

    def home(self, skip_if_homed:bool=False) -> None:
        if not skip_if_homed:
            self.send_command(COMMANDS["home"])
        else:
            if not self.is_homed():
                self.send_command(COMMANDS["home"])
            else:
                logging.info("Skipping homing")

    def get_dimensions(self) -> None:
        self.send_command(COMMANDS["get_dimensions"])

    def abort(self) -> None:
        self.send_command(COMMANDS["abort"])

    def retract(self) -> None:
        self.send_command(COMMANDS["retract"])

    #Retracts the spatula if necessary, spins to stack id and raises spatula for plate to be placed
    def load(self, stack_id:int, plate_height:float, plate_thickness:float, plate_stack_height:float) -> None:
        if stack_id > 16:
            raise RuntimeError("Invalid Stack Id. Range must be 0-16")
        self.home(True)
        stack_id -= 1
        self.set_plate_dimensions(plate_stack_height, plate_thickness, plate_height)
        try:
            self.send_command(f"l {stack_id}")
        except Exception:
            raise RuntimeError(f"Failed to Load plate into {stack_id}")
        self.set_plate_dimensions(plate_stack_height, plate_thickness, plate_height)
        print(f"Loaded plate into stack {stack_id}")

    #Presents the plate to be unloaded by the robot arm
    def unload(self, stack_id:int, plate_height:float, plate_thickness:float, plate_stack_height:float) -> None:
        if stack_id > 16:
            raise RuntimeError("Invalid Stack Id. Range must be 0-16")
        self.home(True)
        stack_id -= 1
        self.set_plate_dimensions(plate_stack_height, plate_thickness, plate_height)
        try:
            self.send_command(f"u {stack_id}")
        except Exception:
            raise RuntimeError(f"Failed to unload plate from {stack_id}")
        self.set_plate_dimensions(plate_stack_height, plate_thickness, plate_height)
        print(f"Unloaded plate from stack {stack_id}")

    def go_to(self, stack_id:int) -> None:
        if stack_id > 16:
            raise RuntimeError("Invalid Stack Id. Range must be 0-16")
        stack_id -= 1
        logging.info(f"Spinning to stack {stack_id}")
        cmd = COMMANDS["spin"]
        self.send_command(f"{cmd} {stack_id}", 30000)

    def set_plate_dimensions(self, plate_height:float, stack_height:float, plate_thickness:float) -> None:
        logging.info("Setting microserve plate dimensions")
        logging.info(f"Plate height= {plate_height}")
        logging.info(f"Stacked height= {stack_height}")
        logging.info(f"Well Height= {plate_thickness}")
        cmd = COMMANDS["set_plate_dimensions"]
        self.send_command(f"{cmd} {plate_height*1000} {stack_height*1000} {plate_thickness*1000}", 5000)

    def set_to_manual(self) -> None:
        self.send_command(COMMANDS["manual"])

    def disconnect(self) -> None:
        if self.tcp:
            self.tcp.disconnect()

    def send_command(self, message:str, timeout:int=30000) -> str:
        if not self.tcp:
            raise ConnectionError("Not connected to any server.")
        self.tcp.clear_buffer()
        result = self.tcp.send_command(message)
        logging.info(f"Acknowledgement response is {result}")
        if "ACK!" not in result:
            raise ValueError(f"Invalid response: {result}")
        
        wait_response = self.wait_for_command(timeout)
        return wait_response

    def wait_for_command(self, timeout:int) -> str:
        response = ""
        if self.blocking and self.tcp:
            response = self.tcp.read_response(timeout=timeout)
            if response.lower().startswith(("error", "aborted")):
                raise ValueError(f"Failed to complete command. Error: {response}")
        return response
    
if __name__ == "__main__":
    microserve = MicroServeDriver(ip="192.168.1.60", port=1000)
    microserve.connect()
    microserve.is_homed()
    microserve.get_status()
    microserve.send_command("err")
    # microserve.get_dimensions()
    # microserve.set_plate_dimensions(15000,15000,13000)
    # #microserve.is_homed()
    # microserve.home()
    # print("Microserve homed")
    # print("Going to 3")
    # microserve.go_to(8)
    # # microserve.unload(3)
    # print("Microserve went to 8")
    # microserve.unload(8)
    # print("Microserve unload plate")
    # microserve.load(8)
    # print("Microserve loaded plate ")
    # microserve.set_to_manual()
    # print("Microserve is in manual mode")
    # print("Microserve went to 5")
    # microserve.go_to(1)
    # print("Microserve went to 1")
    # microserve.go_to(7)
    # print("Microserve went to 7")



import os 



# def kill_processes(self, process_name: str) -> None:
#         # check if any processes are running (length of returned list is > 0)
#         if len([proc_str for proc_str in os.popen('tasklist').readlines() if proc_str.startswith(process_name)]) > 0:

#             # # ask user with pop whether or not to kill processes
#             # style_flag = 0x00001124 # 1 -> sys modal,no; 1 -> default to NO button; 2 -> add ? icon, 4 -> yes/no button style
#             # return_val = ctypes.windll.user32.MessageBoxW(0, f"Kill running {process_name} processes?", f"Kill {self.__class__} Processes", style_flag)
            
#             # # if "OK" then kill 'em
#             # if return_val == 1:
#             #     logging.info("ABCToolDriver Killing Previously Active %s processes", process_name)
#             #     while len([proc_str for proc_str in os.popen('tasklist').readlines() if proc_str.startswith(process_name)]) > 0:

#             #         # kill process(es)
#             #         os.system(f"taskkill /f /im {process_name}")

for proc_str in os.popen('tasklist').readlines():
    print(proc_str)


from google.protobuf.struct_pb2 import Struct
import typing as t

def struct_to_dict(struct: Struct) -> t.Any:
    out = {}
    for key, value in struct.items():
        if isinstance(value, Struct):
            out[key] = struct_to_dict(value)
        else:
            out[key] = value
    return out



import subprocess 
import os 
import logging 
import typing as t

def write_to_file(file_name: str, content: str) -> None:
    with open(file_name, "w") as f:
        f.write(content)

def run_python_script(script_content: str, blocking: bool = True) -> t.Optional[str]:
    #Create a temporary file to run the script
    if os.path.exists("tmp_file.py"):
        os.remove("tmp_file.py")
    if os.path.exists("stdout.txt"):
        os.remove("stdout.txt")
    temp_file = "tmp_file.py"
    #script_content = script_content.encode("utf-8").decode("unicode_escape")
    # Use a raw string to avoid escape sequence issues
    # This preserves all whitespace and newlines as they are
    script_content = r"{}".format(script_content)
    write_to_file(temp_file, script_content)

    if not os.path.exists(temp_file):
        raise RuntimeError("Invalid file path")
    cmd = ["python", "-m", temp_file.replace(".py", "").lstrip("/").replace("/", ".")]
    logging.info("Command: " + str(cmd))
    try:
        process = subprocess.Popen(cmd, stdout=open('stdout.txt', 'w'),stderr=subprocess.STDOUT)
        if blocking:
            process.wait()
            if process.returncode != 0:
                with open('stdout.txt', 'r') as f:
                    error_output = f.read()
                raise RuntimeError(f"Script failed with return code {process.returncode}. Output:\n{error_output}")
            else:
                with open('stdout.txt', 'r') as f:
                    return f.read()
    except FileNotFoundError:
        logging.error("Python executable not found.")
        raise
    except subprocess.CalledProcessError as e:
        logging.error(f"There was an error while running script: {e}")
        raise
    finally:
        logging.info("Cleaning up temporary files")
        os.remove(temp_file)
        os.remove('stdout.txt')
    return None


if __name__ == "__main__":
    print(run_python_script('print("Hello Worldaaaa")\nfor i in range(0,40):print(i)',blocking=True)) 


import logging 
import os
from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.hig_centrifuge_pb2 import Command, Config
from .driver import HiGCentrifugeDriver
import argparse 

class HiGCentrifugeServer(ToolServer):
    toolType = "hig_centrifuge"
    driver: HiGCentrifugeDriver
    config: Config

    def __init__(self) -> None:
        super().__init__()
        self.driver: HiGCentrifugeDriver
    
    def _configure(self, request:Config) -> None:
        logging.info("Initializing HiG")
        self.config = request
        self.driver = HiGCentrifugeDriver(can_port=0)
        self.driver.initialize()
        self.driver.home()
        
    def Home(self, params: Command.Home) -> None:
        self.driver.home()
    
    def Spin(self, params: Command.Spin) -> None:
        self.driver.spin(params.speed, params.acceleration, params.decceleration, params.duration)
    
    def OpenShield(self, params: Command.OpenShield) -> None:
        self.driver.open_shield(params.bucket_id)
    
    def CloseShield(self, params: Command.CloseShield) -> None:
        self.driver.close_shield()
    
    def EstimateHome(self, params: Command.Home) -> int:
        return 1
    def EstimateSpin(self, params: Command.Spin) -> int:
        return 1
    def EstimateOpenShield(self, params: Command.OpenShield) -> int:
        return 1
    def EstimateCloseShield(self, params: Command.CloseShield) -> int:
        return 1

    

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
         raise RuntimeWarning("Port must be provided...")
    serve(HiGCentrifugeServer(),os.environ.get("PORT",str(args.port)))


import serial
import time
import logging
import typing as t

from tools.base_server import ABCToolDriver

ERROR_CODES = {
            "101": "DC motor controller error.",
            "102": "Error caused by speed failure (e.g mechanical locking).",
            "103": "Shaker not initialized or initialization parameters were incorrect after powering on.",
            "104": "Initialization error.",
            "105": "Shaker did not home after stop command.",
            "201": "Error due to failed response from temperature sensors or incorrect internal settings of temperature sensors.",
            "202": "Temperature communication bus error.",
            "203": "Sensor with the requested ID was not found.",
            "204": "Errors caused by a faulty temperature measurement during run.",
            "300": "General error.",
            "301": "IC-Driver error.",
            "303": "Verification error by the unlock position. Please RESTART the system.",
            "304": "Error caused by unsuccessful reach the lock position (timeout). Please RESTART the system.",
            "305": "Error caused by unsuccessful reach the unlock position (timeout). Please RESTART the system.",
            "306": "Error caused by unsuccessful reach the lock position (over current). Please RESTART the system.",
            "307": "Error caused by unsuccessful reach the unlock position (over current). Please RESTART the system.",
        }

SHAKE_STATES = {
                "0": "Shaking",
                "1": "Shake Cycle Stopped and Homed",
                "2": "Shake Cycle Emergency Stop. Re-home shaker",
                "3": "Home",
                "4": "Manual Mode",
                "5": "Accelerating",
                "6": "Decelerate",
                "7": "Decelerate to Stop.",
                "90": "ECO Mode.",
                "99": "Boot Process Running.",
            }

class BioshakeDriver(ABCToolDriver):
    def __init__(self, port: str):
        self.port = port
        self.ser = serial.Serial(port, baudrate=9600, timeout=3)
        self.shake_state = None
        self.connect()

    def __del__(self) -> None:
        self.disconnect()

    def _send_command(self, command: str) -> str:
        if not self.ser.is_open:
            self.ser.open()

        full_command = command + "\r\n"
        logging.debug(f"Sending command: {command}")
        self.ser.write(full_command.encode("ascii"))

        response = self.ser.readline().decode("ascii").strip()
        logging.debug(f"Received response: {response}")

        if response == "e":
            error_list = self.get_error_list()
            raise Exception("BioShake error: " + error_list)
        return response

    def get_error_list(self) -> str:
        result = self._send_command("gel").strip()
        errors = []
        for err in result.split(";"):
            code = err.strip().strip("{}")
            if code in ERROR_CODES:
                errors.append(ERROR_CODES[code])
            else:
                errors.append(code)
        error_message = "\n".join(errors)
        logging.error("Error list: " + error_message)
        return error_message

    def connect(self) -> None:
        if not self.ser.is_open:
            self.ser.open()
        else:
            self.reset_device()
            self.ser.close()
            self.ser.open()

    def disconnect(self) -> None:
        if self.ser.is_open:
            self.ser.close()

    def _is_clamp_locked(self) -> bool:
        elm_state = self.get_elm_state_as_string()
        if elm_state.lower() == "elmlocked":
            return True
        else:
            return False

    def grip(self) -> None:
        self._send_command("selp")

    def ungrip(self) -> None:
        if self._get_shake_state() != "Home":
            self.home()
            # Wait a bit to be sure the device has homed.
            while self._get_shake_state() != "Home":
                time.sleep(0.1)
        self._send_command("seup")

    def home(self) -> None:
        if not self._is_clamp_locked():
            self.grip()
        self._send_command("sgh")
        while self._get_shake_state() != "Home":
            time.sleep(0.1)

    def start_shake(self, speed: int, duration: int) -> None:
        if not self._is_clamp_locked():
            self.grip()

        current_speed_str = self._send_command("gsts")
        try:
            current_speed = float(current_speed_str)
        except ValueError:
            logging.error("Invalid shake speed received.")
            return

        if int(current_speed) == 0:
            logging.error(f"Shake speed is: {current_speed}. Please set to a positive nonzero value!")
        else:
            self._send_command("son")
        # (Duration handling could be added here if desired.)

    def stop_shake(self) -> None:
        self._send_command("soff")

    def reset(self) -> None:
        self._send_command("reset")

    def wait_for_shake(self) -> None:
        while self._get_shake_state() != "Home":
            time.sleep(0.1)

    def _set_shake_speed(self, speed: int) -> None:
        if 0 < speed < 999999:
            self._send_command("ssts" + str(speed))
        else:
            logging.error("Please enter a valid target speed. (Non-negative and no larger than 6 digits.)")

    def _set_acceleration(self, acceleration: int) -> None:
        if 0 < acceleration <= 99:
            self._send_command("ssa" + str(acceleration))
        else:
            logging.error("Please enter an acceleration value within a 1 to 2 digit range.")

    def _get_remaining_time(self) -> int:

        response = self._send_command("gsrt")
        try:
            return int(response)
        except ValueError:
            logging.error("Invalid remaining time received.")
            return 0

    def _get_shake_state(self) -> str:

        result = self._send_command("gsst")
        if result in SHAKE_STATES:
            return SHAKE_STATES[result]
        elif result is None:
            return "Unable to return Shaker status"
        return result

    def _get_target_speed(self) -> int:
        response = self._send_command("gsts")
        try:
            return int(response)
        except ValueError:
            logging.error("Invalid target speed received.")
            return 0

    def _get_speed_actual(self) -> int:
        response = self._send_command("gsas")
        try:
            return int(response)
        except ValueError:
            logging.error("Invalid actual speed received.")
            return 0

    def _get_acceleration_actual(self) -> int:
        response = self._send_command("gsa")
        try:
            return int(response)
        except ValueError:
            logging.error("Invalid acceleration value received.")
            return 0

    def temp_on(self) -> None:
        self._send_command("ton")

    def temp_off(self) -> None:
        self._send_command("toff")

    def _set_temp(self, temp: int) -> None:
        if 0 < temp < 990:
            self._send_command("stt" + str(temp))
        else:
            logging.error(f"Enter a valid temperature target {temp/10}. Must be a three digit number in the range 0-99.")

    def get_elm_state_as_string(self) -> str:
        return self._send_command("gesas")

    def set_elm_lock_pos(self) -> None:
        self.grip()

    def set_elm_unlock_pos(self) -> None:
        self.ungrip()

    def set_shake_target_speed(self, rpm: int) -> None:
        self._set_shake_speed(rpm)

    def shake_on_with_runtime(
        self, seconds: int, speed: t.Optional[int] = None, acceleration: t.Optional[int] = None
    ) -> None:
        if not self._is_clamp_locked():
            self.grip()
        if speed is not None:
            self._set_shake_speed(speed)
        if acceleration is not None:
            self._set_acceleration(acceleration)
        else:
            self._set_acceleration(10)
        self._send_command("sonwr" + str(seconds))
        while self._get_shake_state() != "Home":
            time.sleep(0.1)
        self.ungrip()

    def get_shake_remaining_time(self) -> int:
        return self._get_remaining_time()

    def shake_off(self) -> None:
        self.stop_shake()

    # For compatibility with the blank template.
    def shake_with_time(self, duration: int, speed: int) -> None:
        self.shake_on_with_runtime(seconds=duration, speed=speed)


if __name__ == "__main__":
    try:
        logging.basicConfig(
            level=logging.DEBUG,  # Set the logging level (DEBUG, INFO, etc.)
            format="%(asctime)s - %(levelname)s - %(message)s",
        )
        driver = BioshakeDriver(port="COM7")
        driver.connect()
        print("version: ", driver.get_version())
        print("ELM state: ", driver.get_elm_state_as_string())
        print("locking...")
        driver.set_elm_lock_pos()
        print("Shake state: ", driver._get_shake_state())
        print("setting target speed to 200...")
        driver.set_shake_target_speed(rpm=200)
        print("start shaking for 10 seconds...")
        driver.shake_on_with_runtime(seconds=10)
        for i in range(5):
            print("Remaining time: ", driver.get_shake_remaining_time())
            time.sleep(1)
        print("stop shaking...")
        driver.shake_off()
        shake_state = driver._get_shake_state()
        print("Shake state: ", shake_state)
        if shake_state == "ELMLocked":
            print("unlocking...")
            driver.set_elm_unlock_pos()
        print("resetting device...")
        driver.reset()
    finally:
        driver.disconnect()



from telnetlib import Telnet
import logging
from typing import Union

def try_utf_decode(data:Union[str,bytes]) -> str:
    if isinstance(data, str):
        return data
    data_string = ""
    try:
        data_string = data.decode("utf-8")
    except Exception:
        raise Exception(f"error decoding to utf8 string: {data!r}")
    return data_string


def telnet_read(conn: Telnet, timeout: int=1) -> str:
    reply_string = conn.read_until(b"\r\n", timeout=timeout)
    reply = try_utf_decode(reply_string)

    if len(reply) < 2 or reply[-1] != "\n" or reply[-2] != "\r":
        raise Exception(f"Received invalid message {reply} from tcp connection.")

    logging.info(f"Received {reply[0:-2]}")

    return reply[0:-2]

class Pf400TcpIp:
    def __init__(self, tcp_host: str, tcp_port: int) -> None:
        self.conn = None
        try:
            self.conn = Telnet(tcp_host, tcp_port, timeout=1)
            self.write_and_read("mode")
        except Exception as e:
            logging.error(f"Failed to establish connection: {e}")
            if self.conn:
                try:
                    self.conn.close()
                except Exception as close_error:
                    logging.warning(f"Error closing connection: {close_error}")
                self.conn = None
            raise

    # PF400 should always return a single line of output, unless you are in "pc" mode.
    def read(self, timeout: int=5) -> str:
        if not self.conn:
            raise Exception("No active connection")
        return telnet_read(self.conn, timeout)

    def read_all(self) -> list[str]:
        if not self.conn:
            raise Exception("No active connection")
        messages = []

        more = True

        while more:
            try:
                msg = self.read(timeout=1)
                messages.append(msg)
            except Exception:
                more = False

        return messages

    def write(self, msg: str) -> None:
        if not self.conn:
            raise Exception("No active connection")
        logging.info(f"Sending {msg}")
        self.conn.write((msg + "\n").encode("utf-8"))

    def write_and_expect(self, msg: str, expected: str="0", timeout: int=5) -> None:
        if not self.conn:
            raise Exception("No active connection")
        command_name = msg.split(" ")[0]

        self.write(msg)
        result = self.read(timeout)

        if result != expected:
            raise Exception(f"Robot returned {result} for {command_name}. Expected {expected}")

    # Does NOT validate return value.
    def write_and_read(self, msg: str, timeout: int=5) -> str:
        if not self.conn:
            raise Exception("No active connection")
        self.write(msg)
        return self.read(timeout)

    def wait_for_eom(self) -> None:
        if not self.conn:
            raise Exception("No active connection")
        self.write("waitForEom")
        result = self.read(timeout=150)

        if result != "0":
            raise Exception(f"Robot returned {result} for waitForEom")

    def close(self) -> None:
        if self.conn:
            try:
                self.write_and_read("attach 0")
                self.conn.close()
            except Exception as e:
                logging.warning(f"Error closing connection: {e}")
            self.conn = None



import os
import logging 
import threading 
import time
from tools.base_server import ABCToolDriver
from typing import Union

try:
    import pythoncom
except Exception:
    # The driver will error if there's no pythoncom
    pass

if os.name == "nt":
    import clr  # type: ignore
    SDK_DLL = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "deps",
                "AxInterop.PlateLocLib.dll",
            )

    clr.AddReference(SDK_DLL) 

    from AxPlateLocLib import AxPlateLoc  # type: ignore

else:
    import enum 

    class AxPlateLoc():  # type: ignore
        ExportAsFormat = enum.Enum("ExportAsFormat", ["COLUMNS"])
        def __init__(self) -> None:
            pass
    
        def Initialize(self) -> None:
            raise NotImplementedError(
                "AxPlateLoc is not supported on non-Windows platforms"
            )
        
        def Close(self) -> None:
            raise NotImplementedError(
                "AxPlateLoc is not supported on non-Windows platforms"
            )
        
        def SetSealingTime(self) -> None:
            raise NotImplementedError(
                "AxPlateLoc is not supported on non-Windows platforms"
            )
        
        def SetSealingTemperature(self) -> None:
            raise NotImplementedError(
                "AxPlateLoc is not supported on non-Windows platforms"
            )
        def GetActualTemperature(self) -> None:
            raise NotImplementedError(
                "AxPlateLoc is not supported on non-Windows platforms"
            )
        def GetSealingTime(self) -> None:
            raise NotImplementedError(
                "AxPlateLoc is not supported on non-Windows platforms"
            )       
        def MoveStageIn(self) -> None:
            raise NotImplementedError(
                "AxPlateLoc is not supported on non-Windows platforms"
            )
        def MoveStageOut(self) -> None:
            raise NotImplementedError(
                "AxPlateLoc is not supported on non-Windows platforms"
            )
        def GetLastError(self) -> None:
            raise NotImplementedError(
                "AxPlateLoc is not supported on non-Windows platforms"
            )
        def ShowDiagsDialog(self) -> None:
            raise NotImplementedError(
                "AxPlateLoc is not supported on non-Windows platforms"
            )        
class PlateLocDriver(ABCToolDriver):
    def __init__(self , profile:str) -> None:
        self.profile: str = profile
        self.live : bool  = False
        self.client :AxPlateLoc
        self.lock = threading.Lock()  # To synchronize access to the device
        self.instantiate()
        self.initialize()


    def instantiate(self) -> None:
        pythoncom.CoInitialize()
        self.client = AxPlateLoc()
        self.client.CreateControl()
        self.client.Blocking = True
        pythoncom.CoUninitialize()

    def initialize(self) -> None: 
        args : dict = {"profile":self.profile}
        self.execute_command("initialize", args)
    
    def close(self) -> None:
        self.schedule_threaded_command("close", {})
    
    def seal(self) -> None:
        self.schedule_threaded_command("seal", {})
        return 
    
    def set_seal_time(self, time:float) -> None:
        self.schedule_threaded_command("set_seal_time", {"time":time})
        return 

    def set_temperature(self, temperature:float) -> None:
        self.schedule_threaded_command("set_temperature", {"temperature":temperature})
        return 
    
    def get_actual_temperature(self) -> None:
        self.schedule_threaded_command("get_actual_temperature", {})
        return 

    def stage_in(self) -> None:
        self.schedule_threaded_command("stage_in", {})
        return 
    
    def stage_out(self) -> None:
        self.schedule_threaded_command("stage_out", {})
        return 
    
    def show_diagnostics(self) -> None:
        self.schedule_threaded_command("show_diagnostics",{})

    def schedule_threaded_command(self, command:str, arguments:dict) -> None:  # type: ignore
        self.execution_thread = threading.Thread(target=self.execute_command(command, arguments,)) # type: ignore
        self.execution_thread.daemon = True
        self.execution_thread.start()
        return None

    def execute_command(self, command:str, arguments:dict) -> None:
        response : Union[int, tuple] = 0
        try:
            if command == "initialize":
                response = self.client.Initialize(arguments["profile"])
            elif command == "seal":
                response = self.client.ApplySeal()
            elif command == "close":
                response  = self.client.Close()
            elif command == "set_seal_time":
                response = self.client.SetSealingTime(arguments["time"])
            elif command == "set_temperature":
                response=  self.client.SetSealingTemperature(arguments["temperature"])
            elif command == "get_actual_temperature":
                response = self.client.GetActualTemperature()
            elif command == "get_seal_time":
                response = self.client.GetSealingTime()
            elif command == "stage_in":
                response = self.client.MoveStageIn()
            elif command == "stage_out":
                response = self.client.MoveStageOut()
            elif command == "get_last_error":
                response = self.client.GetLastError()
            elif command == "show_diagnostics":
                response = self.client.ShowDiagsDialog(arguments["modal"],arguments["level"])
            else:
                response = -1
        except RuntimeError as e:
            self.live = False
            raise RuntimeError(f"Failed to execute command {str(e)}")
        finally:
            time.sleep(1)
            pythoncom.CoUninitialize()
            logging.info(f"Received response is {response}")
            logging.info(f"Reponse type is {type(response)}")
            error : str = ""
            if isinstance(response, tuple):
                if response[0] != 0:
                    error = self.client.GetLastError()
                    raise RuntimeError(f"Failed to execute command {error}")
            elif isinstance(response, int):
                if response != 0:
                    error = self.client.GetLastError()
                    raise RuntimeError(f"Failed to execute command {error}")
            return None


##Non threaded commands work fine except via grpc, still not sure why but this is the only way I made it work. Eg. repace schedule_threaded_command with execute_command
if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    driver = PlateLocDriver("plateLoc")
    driver.initialize()
    driver.set_seal_time(2.5)
    # driver.set_temperature()
    # driver.get_actual_temperature()
    # driver.stage_in()
    # driver.stage_out()
    # driver.stage_in()
    # driver.stage_out()
    # driver.show_diagnostics()
    # #driver.close()